/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#include <iostream>     // cout,cerr
#include <fstream>      // ifstream
#include <sysexits.h>
#include "Arguments.h"

using namespace std;

ProgramArguments programArguments;

std::string failureTypeName[] = { "NO_FAILURES", "CRASH_AT_START", "RANDOM_CRASH" };


void usage(void)
{
	cerr << "\n\
Program arguments must be specified in a text file.\n\
Options should be listed one per line in 'name format' value.\n\
Lines starting with '#' and blank lines are ignored.\n\
\n\
Possible options (default value in brackets):\n\
\n\
\tNumProcessesMin: minimum number of processes   [NO DEFAULT]\n\
\tNumProcessesMax: maximum number of processes   [NO DEFAULT]\n\
\tNumXPoints: number of points in the X axis   [10]\n\
\tNumSimulations: number of repetitions to simulate   [100]\n\
\tpfMin: minimum percentage of failed processes   [0]\n\
\tpfMax: maximum percentage of failed processes   [50]\n\
\tpfStep: step increment for the percentage of failed processes   [10]\n\
\tFailureType: how to simulate process failures   [1]\n\
\t\t0 = do not fail.\n\
\t\t1 = crash just before starting.\n\
\t\t2 = crash at a random point during execution.\n\
\tDebug: debug level (0=no, >0 more verbose)   [0]\n\
\tShowProgress: show progress indication (on stderr) 0=no, 1=yes   [1]\n\
\tPlotSubtitle: text to add to the plot subtitle   []\n\
\n\
Additionally, all text between a line with '>>>' and another line with '<<<'\n\
will be treated as comments to be added to the R files.\n\
\n\
\n\
This program runs simulations with a number of processes (n) between NumProcessesMin and NumProcessesMax, so that NumPoints approximately equally spaced cases are simulated.\n\
\n\
In each case, NumSimulations executions are performed, and results are aggregated.\n\
\n\
In each case, the number of failed processes varies from 0% to 50%, in 10% increments.\n\
\n\
\n\
\n";
}


static bool analyzeArgumentsLine(const string& line, string& name, string& value, string& error)
{
	size_t pos1, pos2;

	error = "";
	pos1 = line.find_first_not_of(" \t");
	if (pos1 == string::npos || line[pos1] == '#')
        return false;
    pos2 = line.find_first_of(" \t", pos1);
    if (pos2 == string::npos) {
        error = "Syntax error. Separator is missing.";
        return true;
    }
    name = line.substr(pos1, pos2-pos1);
    pos1 = line.find_first_not_of(" \t", pos2);
    if (pos1 == string::npos) {
        error = "Syntax error. Value is missing.";
        return true;
    }
    value = line.substr(pos1);
    return true;
}


void readArguments(int argc, char** argv, ProgramArguments& programArguments)
{
	string line, name, value;
	unsigned int numLine = 0;
	string error;
	bool inComment;

	if (argc != 2) {
		usage();
		exit(EX_USAGE);
	}
	ifstream fileArguments(argv[1]);
	if (! fileArguments.is_open()) {
        cerr << argv[1] << ": error opening the argument file.\n";
        exit(EX_NOINPUT);
	}
	inComment = false;
	while (getline(fileArguments, line)) {
        numLine++;
        if (inComment) {
            if (line == "<<<")
                inComment = false;
            else {
                programArguments.comments.append(line);
                programArguments.comments.append("\n# ");
            }
        }
        else {
            if (line == ">>>") {
                inComment = true;
                if (programArguments.comments.size() == 0)
                    programArguments.comments = "\n\n# ";
            }
            else {
                if ( ! analyzeArgumentsLine(line, name, value, error))
                    continue;
                if (error.size() > 0) {
                    cerr << "Line " << numLine << ": " << error << '\n';
                    exit(EX_DATAERR);
                }
                if (name == "NumProcessesMin")
                    programArguments.nMin = atoi(value.c_str());
                else if (name == "NumProcessesMax")
                    programArguments.nMax = atoi(value.c_str());
                else if (name == "NumXPoints")
                    programArguments.numXPoints = atoi(value.c_str());
                else if (name == "NumSimulations")
                    programArguments.numSimulations = atoi(value.c_str());
                else if (name == "pfMin")
                    programArguments.pfMin = atoi(value.c_str());
                else if (name == "pfMax")
                    programArguments.pfMax = atoi(value.c_str());
                else if (name == "pfStep")
                    programArguments.pfStep = atoi(value.c_str());
                else if (name == "FailureType")
                    programArguments.failureType = atoi(value.c_str());
                else if (name == "Debug")
                    programArguments.debugLevel = atoi(value.c_str());
                else if (name == "ShowProgress")
                    programArguments.showProgress = (atoi(value.c_str()) != 0);
                else if (name == "OutputFilename")
                    programArguments.outputFilename = value;
                else if (name == "PlotSubtitle")
                    programArguments.plotSubtitle = value;
                else {
                    cerr << "Line " << numLine << ": unknown option (" << name << ")\n";
                    exit(EX_DATAERR);
                }
            }
        }
	}
	fileArguments.close();
    programArguments.comments.append("\n\n");
	if (programArguments.nMin == 0) {
        cerr << "ERROR: the minimum number of processes is missing.\n";
        exit(EX_DATAERR);
	}
	if (programArguments.nMax == 0) {
        cerr << "ERROR: the maximum number of processes is missing.\n";
        exit(EX_DATAERR);
	}
	if (programArguments.nMax < programArguments.nMin) {
        cerr << "ERROR: the minimum number of processes must be lower than or equal to the maximum number of processes.\n";
        exit(EX_DATAERR);
	}
	if (programArguments.pfMax < programArguments.pfMin) {
        cerr << "ERROR: the minimum percentage of failed processes must be lower than or equal to the maximum percentage of failed processes.\n";
        exit(EX_DATAERR);
	}
	if (programArguments.pfMax > 50) {
        cerr << "ERROR: the percentage of failed processes cannot be higher than 50%.\n";
        exit(EX_DATAERR);
	}
	if (programArguments.pfMin != programArguments.pfMax && programArguments.pfStep == 0) {
        cerr << "ERROR: the step increment for the percentage of failed processes must be higher that zero.\n";
        exit(EX_DATAERR);
	}
	if (programArguments.outputFilename.size() == 0) {
        cerr << "ERROR: the output filename is missing.\n";
        exit(EX_DATAERR);
	}
}
