/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#include "Channel.h"


#include <sstream>
#include <iostream>


_MessageFilterAll MessageFilterAll;


Message::Message(int _type) :
    type(_type)
{
}


Message::~Message()
{
}


MessageFilter::~MessageFilter()
{
}


_MessageFilterAll::~_MessageFilterAll()
{

}


bool _MessageFilterAll::operator()(std::shared_ptr<const Message> message)
{
    return true;
}


ChannelStatistics::ChannelStatistics()
{
}


ChannelStatistics& ChannelStatistics::operator += (const ChannelStatistics& other)
{
    this->numMessagesSent += other.numMessagesSent;
    return *this;
}


std::ostream& operator << (std::ostream& out, const ChannelStatistics& statistics)
{
    std::stringstream s;

    s << "No. messages sent: " << statistics.numMessagesSent << '\n';
    out << s.str();
    return out;
}


Channel::Channel() :
    _debugLevel(0)
{
}


Channel::~Channel()
{
    for (auto const& receiver : messageQueues) {
        freeMessages(receiver.second);
        delete receiver.second;
    }
}


void Channel::debug(int level)
{
	_debugLevel = level;
}


void Channel::attach(int idProcess)
{
    std::lock_guard<std::mutex> l0(generalMutex);
    VectorClock r;

    messageQueues[idProcess] = new MessageList;
    clocks[idProcess] = r;
    breakSignaled[idProcess] = false;
    // Reset all vector clocks to zero
    for (auto const& r : clocks)
        clocks[r.first].resize(clocks.size());
}


void Channel::broadcast(int idProcess, std::shared_ptr<Message> message, bool fail)
{
    std::lock_guard<std::mutex> l0(generalMutex);
    VectorClock& clock = clocks[idProcess];
    int probReorder, probReorderBase;
    unsigned int failureProbability;  // Out of 1000
    unsigned int numMessagesSent;

    clock[idProcess]++;
    message->timestamp(clock);
    probReorder = probReorderBase = 2;
    // If this broadcast has to fail, it will do so by sending just a percentage
    // of the messages
    if (fail)
        failureProbability = random() % 1000;
    numMessagesSent = 0;
    for (auto const& mq : messageQueues) {
        MessageList* receiver = mq.second;
        probReorder = probReorderBase;
         // If a failure was requested it will not send the message if either:
         //   - A random value tells to fail
         //   - All messages but one have been sent (therefore, it's the last one and it must not be sent)
        if (! fail || ((random() % 1000) >= failureProbability && numMessagesSent < messageQueues.size() - 1)) {
            MessageList::reverse_iterator m, m2;
            bool stopReorder = false;
            int maxJump = receiver->size(); // Note: effectively, this puts no limit to the maximum jump a message can do.

            m = receiver->rbegin();
            m2 = m;
            while (! stopReorder && m != receiver->rend() && maxJump-- >= 0) {
                if (_debugLevel >= 20) {
                    std::stringstream s;
                    s << "m-r=" << (*m)->type << ",b=";
                    if (m.base() == receiver->begin())
                        s << "begin\n";
                    else if (m.base() == receiver->end())
                        s << "end\n";
                    else
                        s << (*(m.base()))->type << '\n';
                    if ((*m)->t() < message->t())
                        s << message->type << ">" << (*m)->type << '\n';
                    if (! ((*m)->t() < message->t()) && ! (message->t() < (*m)->t()))
                        s << message->type << "||" << (*m)->type << '\n';
                    std::cerr << s.str();
                }
                if ((*m)->t() < message->t()) {
                    stopReorder = true;
                    if ((random() % probReorder++) == 0)
                        m2 = m;
                }
                else if (! ((*m)->t() < message->t()) && ! (message->t() < (*m)->t()) && (random() % probReorder++) == 0)
                    m2 = m;
                m++;
            }
            if (_debugLevel >= 20) {
                std::stringstream s;
                s << "m2-r=";
                if (m2 == receiver->rbegin())
                    s << "rbegin";
                else if (m2 == receiver->rend())
                    s << "rend";
                else
                    s << (*m2)->type;
                s << ",b=";
                if (m2.base() == receiver->begin())
                    s << "begin\n";
                else if (m2.base() == receiver->end())
                    s << "end\n";
                else
                    s << (*(m2.base()))->type << '\n';
                std::cerr << s.str();
            }
            if (stopReorder || (random() % 2) == 0)
                receiver->insert(m2.base(), message);
            else
                receiver->insert(--m2.base(), message);
            numMessagesSent++;
        } // if (! fail ...
    }
    if (_debugLevel >= 20) {
        std::stringstream s;
        for (auto const& receiver : messageQueues) {
            s << receiver.first << " -> ";
            for (auto const& m : *receiver.second) {
                s << m->type << "[" << m->t() << "] ";
            }
            s << '\n';
        }
        s << "---\n";
        std::cerr << s.str();
    }
    statistics.numMessagesSent.tmpValue += numMessagesSent;
    cvReception.notify_all();
}


std::shared_ptr<Message> Channel::receive(int idProcess, MessageFilter& filter, bool blocking)
{
    std::unique_lock<std::mutex> l0(generalMutex);
    std::shared_ptr<Message> message;
    VectorClock& clock = clocks[idProcess];

    if (breakSignaled[idProcess]) {
        breakSignaled[idProcess] = false;
        return 0;
    }
    message = 0;
    do {
        for (auto const& m : *messageQueues[idProcess]) {
            if (filter(m)) {
                message = m;
                break;
            }
        }
        if (message == 0) {
            if (blocking)
                cvReception.wait(l0);
            else
                return 0;
        }
    } while (message == 0 && ! breakSignaled[idProcess]);
    if (message != 0) {
        clock[idProcess]++;
        clock += message->t();
        messageQueues[idProcess]->remove(message);
        return message;
    }
    else {
        breakSignaled[idProcess] = false;
        return 0;
    }
}


struct RemoveFilter {
    RemoveFilter(MessageFilter& _filter) : filter(_filter) {}
    bool operator() (const std::shared_ptr<const Message>& m) { return filter(m); }
    MessageFilter& filter;
};


void Channel::discard(int idProcess, MessageFilter& filter)
{
    std::unique_lock<std::mutex> l0(generalMutex);
    RemoveFilter f(filter);
    int antes, despues;

    antes = (*messageQueues[idProcess]).size(); // JMR
    (*messageQueues[idProcess]).remove_if(f);
    despues = (*messageQueues[idProcess]).size(); // JMR
    if (despues > 10000) // JMR
    {
        std::stringstream s;
        s << "P" << idProcess << ": MQ.size=" << (*messageQueues[idProcess]).size() << ", borrados " << antes-despues << ", clock=" << clocks[idProcess] << '\n';
        std::cerr << s.str() << std::flush;
    }
}


void Channel::reset()
{
    std::unique_lock<std::mutex> l0(generalMutex);

    for (auto const& receiver : messageQueues) {
        freeMessages(receiver.second);
        receiver.second->clear();
    }
    for (auto& b : breakSignaled)
        b.second = false;
    statistics.numMessagesSent.tmpValue = 0;
}


void Channel::freeMessages(MessageList* messageList)
{
    for (auto& m : *messageList)
        m.reset();
}


void Channel::freeMessageList(MessageList* messageList)
{
    freeMessages(messageList);
    delete messageList;
}


void Channel::breakWait()
{
    std::unique_lock<std::mutex> l0(generalMutex);
    for (auto& b : breakSignaled)
        b.second = true;
    cvReception.notify_all();
}


void Channel::collectStatistics()
{
    std::unique_lock<std::mutex> l0(generalMutex);

    statistics.numMessagesSent.collect();
}


