/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#ifndef __CHANNEL_H__
#define __CHANNEL_H__

#include <memory>
#include <list>
#include <map>
#include <mutex>
#include <condition_variable>
#include "Events.h"
#include "Observations.h"

class Message : public Event {
public:
	Message(int type);
	virtual ~Message();
	const int type;
};

typedef std::list<std::shared_ptr<Message>> MessageList;

class MessageFilter {
public:
    virtual ~MessageFilter();
    virtual bool operator()(std::shared_ptr<const Message> message) = 0;
};


class _MessageFilterAll : public MessageFilter {
public:
    virtual ~_MessageFilterAll();
    bool operator()(std::shared_ptr<const Message> message);
};

extern _MessageFilterAll MessageFilterAll;


struct ChannelStatistics {
    ChannelStatistics();
    ObservedVariable<unsigned int> numMessagesSent;
    ChannelStatistics& operator += (const ChannelStatistics& other);
    friend std::ostream& operator << (std::ostream& out, const ChannelStatistics& statistics);
};


class Channel {
public:
	Channel();
	virtual ~Channel();
	void attach(int idProcess);
	void broadcast(int idProcess, std::shared_ptr<Message> message, bool fail = false);
	std::shared_ptr<Message> receive(int idProcess, MessageFilter& filter, bool blocking);
	void discard(int idProcess, MessageFilter& filter);
	void reset();
    void debug(int level);
    void breakWait();
	static void freeMessages(MessageList* messageList);
	static void freeMessageList(MessageList* messageList);
	void collectStatistics();
    ChannelStatistics statistics;
private:
	int _debugLevel;
	std::condition_variable cvReception;
	std::map<int,VectorClock> clocks; /**< Vector clocks of processes attached to this channel. */
	std::map<int,bool> breakSignaled;
    std::map<int,MessageList*> messageQueues;
	std::mutex generalMutex;
};

#endif
