/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <math.h>
#include <iostream>
#include <sstream>      // stringstream
#include <stdexcept>    // runtime_error
#include <memory>       // std::static_pointer_cast
#include <algorithm>    // std::sort, std::begin, std::end
#include "Coin.h"
#include "Channel.h"
#include "Arguments.h"  // Failure types

using namespace std;

Coin::Coin(unsigned int _numProcesses, unsigned int _maxFailures) :
    numProcesses(_numProcesses),
    maxFailures(_maxFailures),
    _debugLevel(0)
{
}


Coin::~Coin()
{
}


void Coin::debug(int level)
{
	_debugLevel = level;
}

//std::string coinProb0 = "1/n";
std::string coinProb0 = "2/3n";
//std::string coinProb0 = "4/5n";
//std::string coinProb0 = "6/7n";
//std::string coinProb0 = "1.38585/n";
//std::string coinProb0 = "1.1/n";
int Coin::individualToss()
{
    // Prob(0) = 1/n
//	return (random() % numProcesses) == 0 ? 0 : 1;

	// Prob(0) = 2/3n
	return (random() % (3*numProcesses)) < 2 ? 0 : 1;

	// Prob(0) = 4/5n
//	return (random() % (5*numProcesses)) < 4 ? 0 : 1;

	// Prob(0) = 6/7n
//	return (random() % (7*numProcesses)) < 6 ? 0 : 1;

	// Prob(0) = 1.38585/n
	// 1.38585 = 27717/20000
//	return (random() % (20000*numProcesses)) < 27717 ? 0 : 1;

    // Prob(0) = 1.1/n
    // 1.1=11/10
//	return (random() % 10*numProcesses) < 11 ? 0 : 1;
}


void Coin::show(int* values, unsigned int numValues)
{
	for (unsigned int i = 0; i < numValues; i++)
		cout << values[i] << " ";
	cout << '\n';
}


void Coin::shuffle(int* messages, unsigned int numMessages)
{
	for (unsigned int i = 0; i < numMessages; i++) {
		int x, y, aux;
		x = random() % numMessages;
		y = random() % numMessages;
		aux = messages[x];
		messages[x] = messages[y];
		messages[y] = aux;
	}
}


SimpleCoin::SimpleCoin(unsigned int _numProcesses, unsigned int _maxFailures)
	: Coin(_numProcesses, _maxFailures)
{
}


SimpleCoin::~SimpleCoin()
{
}


void SimpleCoin::toss(int* results)
{
	for (unsigned int i = 0; i < numProcesses; i++)
		results[i] = individualToss();
	shuffle(results, numProcesses);
	if (_debugLevel >= 1) {
		cout << "SC  : v[i] = ";
		show(results, numProcesses);
	}
}


CommonToss::CommonToss(unsigned int _numProcesses, unsigned int _maxFailures) :
    numProcesses(_numProcesses),
    maxFailures(_maxFailures),
    numParticipants(1), // The process which starts this toss.
    numPendingConsumers(_numProcesses)
{
    individualTosses = new int[numProcesses];
}


CommonCoin::CommonCoin(unsigned int _numProcesses, unsigned int _maxFailures) :
	Coin(_numProcesses, _maxFailures),
	terminated(false)
{
    numTossProcesses = new unsigned int[numProcesses];
    for (unsigned int i = 0; i < numProcesses; i++)
        numTossProcesses[i] = 0;
}


CommonCoin::~CommonCoin()
{
    delete[] numTossProcesses;
}


void CommonCoin::toss(int* results)
{
	int v[numProcesses];
	int messages[numProcesses];

// Line 1
	for (unsigned int i = 0; i < numProcesses; i++)
		v[i] = individualToss();
	if (_debugLevel >= 5) {
		cout << "START: v[i] = ";
		show(v, numProcesses);
	}
// Line 2
	for (unsigned int i = 0; i < numProcesses; i++)
		messages[i] = v[i];
// Lines 3-6
	for (unsigned int i = 0; i < numProcesses; i++) {
		shuffle(messages, numProcesses);
		v[i] = 1;
		for (unsigned int j = 0; j < numProcesses - maxFailures && v[i] == 1; j++)
			if (messages[j] == 0)
				v[i] = 0;
	}
	if (_debugLevel >= 5) {
		cout << "+PH0  : v[i] = ";
		show(v, numProcesses);
	}
// Line 7
	for (unsigned int i = 0; i < numProcesses; i++)
		messages[i] = v[i];
// Lines 8-11
	for (unsigned int i = 0; i < numProcesses; i++) {
		shuffle(messages, numProcesses);
		v[i] = 1;
		for (unsigned int j = 0; j < numProcesses - maxFailures && v[i] == 1; j++)
			if (messages[j] == 0)
				v[i] = 0;
	}
	if (_debugLevel >= 5) {
		cout << "+PH1  : v[i] = ";
		show(v, numProcesses);
	}
// Line 12
	for (unsigned int i = 0; i < numProcesses; i++)
		messages[i] = v[i];
// Lines 13-16
	for (unsigned int i = 0; i < numProcesses; i++) {
		shuffle(messages, numProcesses);
		v[i] = 1;
		for (unsigned int j = 0; j < numProcesses - maxFailures && v[i] == 1; j++)
			if (messages[j] == 0)
				v[i] = 0;
	}
	if (_debugLevel >= 1) {
		cout << "END : v[i] = ";
		show(v, numProcesses);
	}
	for (unsigned int i = 0; i < numProcesses; i++)
		results[i] = v[i];
}


int CommonCoin::toss(int idProcess, unsigned int failureType)
{
    std::unique_lock<std::mutex> l0(generalMutex);
    unsigned int numToss;
    CommonToss* theToss;
    int result;

    numToss = ++numTossProcesses[idProcess];
    // Looks for this toss or creates it if this is the first process which tosses the coin
    if (tosses.count(numToss) == 0) {
        theToss = new CommonToss(numProcesses, maxFailures);
        tosses[numToss] = theToss;
    }
    else {
        theToss = tosses[numToss];
        theToss->numParticipants++;
    }
    // Wait until there are enough participant processes
    while (! terminated && theToss->numParticipants < theToss->numProcesses-theToss->maxFailures)
        theToss->cvQuorum.wait(l0);
    if (terminated)
        return -1; // TODO: rethink the return value
    // If this is the first process in this toss, generate the toss.
    if (theToss->numPendingConsumers == theToss->numProcesses)
        toss(theToss->individualTosses);
    result = theToss->individualTosses[idProcess];
    // If this is the last process in this toss, delete it.
    if (--theToss->numPendingConsumers == 0) {
        delete theToss;
        tosses.erase(numToss);
    }
    else {
        // Awakes another process.
        theToss->cvQuorum.notify_one();
    }
    return result;
}


void SyntheticCommonCoin::reset()
{
    std::unique_lock<std::mutex> l0(generalMutex);

    numTossProcesses = new unsigned int[numProcesses];
    for (unsigned int i = 0; i < numProcesses; i++)
        numTossProcesses[i] = 0;
    tosses.clear();
    terminated  = false;
}


void SyntheticCommonCoin::terminate()
{
    std::unique_lock<std::mutex> l0(generalMutex);
    // Marks the end of execution and awakes all blocked processes.
    terminated = true;
    for (auto const& tirada : tosses)
        tirada.second->cvQuorum.notify_all();
}


SyntheticCommonCoin::SyntheticCommonCoin(unsigned int _numProcesses, unsigned int _maxFailures, double _bias) :
	CommonCoin(_numProcesses, _maxFailures)
{
    biasRandom = (int) (_bias * rangeRandom);
}


SyntheticCommonCoin::~SyntheticCommonCoin()
{
}


void SyntheticCommonCoin::toss(int* results)
{
    // Produces a common result with the probability stated by the bias parameter.
    if ((random() % rangeRandom) < biasRandom) {
        int valorComun = random() % 2;
        for (unsigned int i = 0; i < numProcesses; i++)
            results[i] = valorComun;
    }
    // Otherwise, produces a non-common result.
    else {
        bool esComun = true;
        do {
            results[0] = random() % 2;
            for (unsigned int i = 1; i < numProcesses; i++) {
                results[i] = random() % 2;
                if (results[i] != results[0])
                    esComun = false;
            }
        } while (esComun);
    }
	if (_debugLevel >= 1) {
		cout << "FINAL : v[i] = ";
		show(results, numProcesses);
	}
}


int SyntheticCommonCoin::toss(int idProcess, unsigned int failureType)
{
    std::unique_lock<std::mutex> l0(generalMutex);
    unsigned int numToss;
    CommonToss* theToss;
    int result;

    if (terminated || failureType == ProgramArguments::FAILURE_START)
        return -1; // TODO: rethink the return value
    numToss = ++numTossProcesses[idProcess];
    // Looks for this toss or creates it if this is the first process which tosses the coin
    if (tosses.count(numToss) == 0) {
        theToss = new CommonToss(numProcesses, maxFailures);
        tosses[numToss] = theToss;
    }
    else {
        theToss = tosses[numToss];
        theToss->numParticipants++;
    }
    // Wait until there are enough participant processes
    while (! terminated && theToss->numParticipants < theToss->numProcesses-theToss->maxFailures)
        theToss->cvQuorum.wait(l0);
    if (terminated)
        return -1; // TODO: rethink the return value
    // If this is the first process in this toss, generate the toss.
    if (theToss->numPendingConsumers == theToss->numProcesses)
        toss(theToss->individualTosses);
    result = theToss->individualTosses[idProcess];
    // If this is the last process in this toss, delete it.
    if (--theToss->numPendingConsumers == 0) {
        delete theToss;
        tosses.erase(numToss);
    }
    else {
        // Awakes another process.
        theToss->cvQuorum.notify_one();
    }
    return result;
}


CommonCoinStatistics::CommonCoinStatistics(unsigned int _n, unsigned int _f) :
    n(_n),
    f(_f),
    numTosses(0),
    numCommonTosses(0),
    xml(false)
{
    for (unsigned int i = 0; i < MAX_V; i++)
        commonValue[i] = 0;
}


CommonCoinStatistics& CommonCoinStatistics::operator += (const CommonCoinStatistics& other)
{
    numTosses += other.numTosses;
    numCommonTosses += other.numCommonTosses;
    for (unsigned int i = 0; i < MAX_V; i++)
        commonValue[i] += other.commonValue[i];
    tossResults.insert(tossResults.end(), other.tossResults.begin(), other.tossResults.end());
    return *this;
}


std::ostream& operator << (std::ostream& out, const CommonCoinStatistics& statistics)
{
    std::stringstream s;

//    std::setprecision(4);
    if (statistics.xml) {
        s << "<common_coin>";
        s << "<num_processes>" << statistics.n << "</num_processes><max_failures>" << statistics.f << "</max_failures>";
        s << "<num_tosses>" << statistics.numTosses << "</num_tosses><num_common_tosses>" << statistics.numCommonTosses << "</num_common_tosses>";
    }
    else {
        s << "COMMON COIN STATISTICS:\n";
        s << "# Processes      : " << statistics.n << '\n';
        s << "Max Failures     : " << statistics.f << '\n';
        s << "# Tosses         : " << statistics.numTosses << '\n';
        s << "# Common Tosses  : " << statistics.numCommonTosses << " (" << (double) statistics.numCommonTosses / statistics.numTosses * 100.0 << "%)\n";
    }

    if (statistics.xml)
        s << "<common_values>";
    for (unsigned int i = 0; i < statistics.MAX_V; i++)
        if (statistics.xml)
            s << "<case><v>" << i << "</v><num>" << statistics.commonValue[i] << "</num></case>";
        else
            s << "# Common Tosses " << i << ": " << statistics.commonValue[i] << " (" << (double) statistics.commonValue[i] / statistics.numCommonTosses * 100.0 << "%)\n";
    if (statistics.xml)
        s << "</common_values>";

    if (statistics.xml)
        s << "</common_coin>\n";
    out << s.str();
    return out;
}


void CommonCoinStatistics::countToss(unsigned int idToss, int value)
{
    int initValue = NOT_TOSSED;

    if (idToss > numTosses)
        numTosses = idToss;
    if (idToss >= tossResults.size())
        tossResults.resize(idToss + INC_SIZE, initValue);
    switch (tossResults[idToss]) {
    case NOT_TOSSED:
        tossResults[idToss] = value;
        commonValue[value]++;
        numCommonTosses++;
        break;
    case NOT_COMMON:
        break;
    default:
        if (tossResults[idToss] != value) {
            tossResults[idToss] = NOT_COMMON;
            numCommonTosses--;
            commonValue[value]--;
        }
    }
}


static int checkUnanimity(int* results, unsigned int n)
{
    bool zeroFound, oneFound;

    zeroFound = (results[0] == 0);
    oneFound = (results[0] == 1);
    for (unsigned int i = 1; i < n; i++) {
        if (results[i] == 0)
            zeroFound = true;
        if (results[i] == 1)
            oneFound = true;
    }
    if (zeroFound && oneFound) {
        std::cerr << "NOT_COMMON" << flush;
        return CommonCoinStatistics::NOT_COMMON;
    }
    else
        return zeroFound ? 0 : 1;
}


void CommonCoinStatistics::countToss(int* results)
{
    int r;

    numTosses++;
    if (numTosses >= tossResults.size()) {
        r = NOT_TOSSED;
        tossResults.resize(numTosses + INC_SIZE, r);
    }
    r = checkUnanimity(results, n);
    tossResults[numTosses] = r;
    if (r != NOT_COMMON) {
        numCommonTosses++;
        commonValue[r]++;
    }
}


MessageCoinPH0::MessageCoinPH0(unsigned int _idToss, int _v) :
    Message(MessageCoinPH0::type),
    idToss(_idToss),
    v(_v)
{
}


MessageCoinPH0::~MessageCoinPH0()
{
}


MessageCoinPH1::MessageCoinPH1(unsigned int _idToss, int _v) :
    Message(MessageCoinPH1::type),
    idToss(_idToss),
    v(_v)
{
}


MessageCoinPH1::~MessageCoinPH1()
{
}


MessageCoinPH2::MessageCoinPH2(unsigned int _idToss, int _v) :
    Message(MessageCoinPH2::type),
    idToss(_idToss),
    v(_v)
{
}


MessageCoinPH2::~MessageCoinPH2()
{
}


class SimulatedCommonCoin::FilterCoinPH : public MessageFilter {
public:
    FilterCoinPH(int messageType, unsigned int idToss);
    bool operator()(std::shared_ptr<const Message> mensaje);
private:
    const int messageType;
    const unsigned int idToss;
};


SimulatedCommonCoin::FilterCoinPH::FilterCoinPH(int _messageType, unsigned int _idToss) :
    messageType(_messageType),
    idToss(_idToss)
{
}


bool SimulatedCommonCoin::FilterCoinPH::operator()(std::shared_ptr<const Message> _mensaje)
{
    if (_mensaje->type == messageType) {
        std::shared_ptr<const MessageCoinPH0> mensaje = std::static_pointer_cast<const MessageCoinPH0>(_mensaje); // TODO: should use the correct message type class
        return idToss==mensaje->idToss;
    }
    else
        return false;
}


class SimulatedCommonCoin::FilterCoinPreviousTosses : public MessageFilter {
public:
    FilterCoinPreviousTosses(unsigned int idToss);
    bool operator()(std::shared_ptr<const Message> mensaje);
private:
    const unsigned int idToss;
};


SimulatedCommonCoin::FilterCoinPreviousTosses::FilterCoinPreviousTosses(unsigned int _idToss) :
    idToss(_idToss)
{
}


bool SimulatedCommonCoin::FilterCoinPreviousTosses::operator()(std::shared_ptr<const Message> _mensaje)
{
    if (_mensaje->type == MessageCoinPH0::type || _mensaje->type == MessageCoinPH1::type || _mensaje->type == MessageCoinPH2::type) {
        std::shared_ptr<const MessageCoinPH0> mensaje = std::static_pointer_cast<const MessageCoinPH0>(_mensaje); // TODO: should use the correct message type class
        return mensaje->idToss <= idToss;
    }
    else
        return false;
}


SimulatedCommonCoin::SimulatedCommonCoin(unsigned int _numProcesses, unsigned int _maxFailures, Channel& _channel) :
	CommonCoin(_numProcesses, _maxFailures),
	statistics(_numProcesses, _maxFailures),
	channel(_channel)
{
}


SimulatedCommonCoin::~SimulatedCommonCoin()
{
}


MessageList* SimulatedCommonCoin::receive(unsigned int idProcess, unsigned int howMany, MessageFilter& filter, std::unique_lock<std::mutex>& coinLock)
{
    MessageList* messageList;

    messageList = new MessageList;
    do {
        coinLock.unlock();
        std::shared_ptr<Message> message = channel.receive(idProcess, filter, true);
        coinLock.lock();
        if (! terminated && message != 0) {
            if (_debugLevel >= 12) {
                std::stringstream s;
                s << "P" << idProcess << ",T" << numTossProcesses[idProcess] << ": received message of type " << message->type << " [" << message->t() << ']' << '\n';
                std::cerr << s.str();
                s.str(std::string());
            }
            messageList->push_back(message);
        }
    } while (messageList->size() < howMany && ! terminated);
    return messageList;
}


void SimulatedCommonCoin::discardUnusedMessages(unsigned int idProcess, std::unique_lock<std::mutex>& coinLock)
{
    // Do not attempt to discard if at round 0. Otherwise, 0-1 = MAXUNSIGNEDINT, and that doesn't make sense.
    if (numTossProcesses[idProcess] > 0) {
        FilterCoinPreviousTosses filter(numTossProcesses[idProcess] - 1);
        std::shared_ptr<Message> message;

        coinLock.unlock();
        channel.discard(idProcess, filter);
    //    while ((message = channel.receive(idProcess, filter, false)) != 0)
    //        message.reset();
        coinLock.lock();
    }
}


int SimulatedCommonCoin::toss(int idProcess, unsigned int failureType)
{
    std::unique_lock<std::mutex> l0(generalMutex);
    unsigned int idToss;
    int result;
    MessageList* messageList;
    unsigned int failPhase;

    idToss = ++numTossProcesses[idProcess];
    // Up to 3*f messages from each toss may not have been used by this process.
    // Discard them, as they are no longer needed and they just make channel simulation slower.
    discardUnusedMessages(idProcess, l0);
    if (terminated)
        return -1; // TODO: rethink the return value
    if (failureType != ProgramArguments::FAILURE_START) {
        // If random failure was requested, it chooses in which phase to fail.
        if (failureType == ProgramArguments::FAILURE_RANDOM)
            failPhase = random() % 3;
        else
            failPhase = 999;
        // Line 1
        result = individualToss();
        if (_debugLevel >= 5) {
            std::stringstream s;
            s << "P" << idProcess << ",T" << idToss << ": individualToss = " << result << '\n';
            std::cerr << s.str();
            s.str(std::string());
        }
        // Line 2
        channel.broadcast(idProcess, std::make_shared<MessageCoinPH0>(idToss, result), failPhase == 0);
        if (_debugLevel >= 5) {
            std::stringstream s;
            s << "P" << idProcess << ",T" << idToss << ": broadcast PH0(v=" << result << ((failPhase == 0) ? ",FAIL" : "") << ")\n";
            std::cerr << s.str();
            s.str(std::string());
        }
        if (failPhase == 0)
            return -1; // TODO: rethink the return value
        // Line 3
        {
            FilterCoinPH filter(MessageCoinPH0::type, idToss);
            messageList = receive(idProcess, numProcesses/2+1, filter, l0);
            if (_debugLevel >= 10) {
                std::stringstream s;
                s << "P" << idProcess << ",T" << idToss << ": received " << messageList->size() << " PH0 messages\n";
                std::cerr << s.str();
                s.str(std::string());
                for (auto const& m : *messageList) {
                    auto m2 = static_pointer_cast<MessageCoinPH0>(m);
                    s << "P" << idProcess << ",T" << idToss << ": PH0(v=" << m2->v << ")\n";
                    std::cerr << s.str();
                    s.str(std::string());
                }
            }
            if (terminated) {
                if (_debugLevel >= 10) {
                    std::stringstream s;
                    s << "P" << idProcess << ",T" << idToss << ": TERMINATE\n";
                    std::cerr << s.str();
                }
                Channel::freeMessageList(messageList);
                return -1; // TODO: rethink the return value
            }
        }
        // Lines 4-6
        {
            bool received0 = false;
            for (auto const& m : *messageList)
                if ((static_pointer_cast<MessageCoinPH0>(m))->v == 0) {
                    received0 = true;
                    break;
                }
            result = received0 ? 0 : 1;
        }
        Channel::freeMessageList(messageList);
        // Line 7
        channel.broadcast(idProcess, std::make_shared<MessageCoinPH1>(idToss, result), failPhase == 1);
        if (_debugLevel >= 5) {
            std::stringstream s;
            s << "P" << idProcess << ",T" << idToss << ": broadcast PH1(v=" << result << ((failPhase == 1) ? ",FAIL" : "") << ")\n";
            std::cerr << s.str();
            s.str(std::string());
        }
        if (failPhase == 1)
            return -1; // TODO: rethink the return value
        // Line 8
        {
            FilterCoinPH filter(MessageCoinPH1::type, idToss);
            messageList = receive(idProcess, numProcesses/2+1, filter, l0);
            if (_debugLevel >= 10) {
                std::stringstream s;
                s << "P" << idProcess << ",T" << idToss << ": received " << messageList->size() << " PH1 messages\n";
                std::cerr << s.str();
                s.str(std::string());
                for (auto const& m : *messageList) {
                    auto m2 = static_pointer_cast<MessageCoinPH1>(m);
                    s << "P" << idProcess << ",T" << idToss << ": PH1(v=" << m2->v << ")\n";
                    std::cerr << s.str();
                    s.str(std::string());
                }
            }
            if (terminated) {
                if (_debugLevel >= 10) {
                    std::stringstream s;
                    s << "P" << idProcess << ",T" << idToss << ": TERMINATE\n";
                    std::cerr << s.str();
                }
                Channel::freeMessageList(messageList);
                return -1; // TODO: rethink the return value
            }
        }
        // Lines 9-11
        {
            bool received0 = false;
            for (auto const& m : *messageList)
                if ((static_pointer_cast<MessageCoinPH1>(m))->v == 0) {
                    received0 = true;
                    break;
                }
            result = received0 ? 0 : 1;
        }
        Channel::freeMessageList(messageList);
        // Line 12
        channel.broadcast(idProcess, std::make_shared<MessageCoinPH2>(idToss, result), failPhase == 2);
        if (_debugLevel >= 5) {
            std::stringstream s;
            s << "P" << idProcess << ",T" << idToss << ": broadcast PH2(v=" << result << ((failPhase == 2) ? ",FAIL" : "") << ")\n";
            std::cerr << s.str();
            s.str(std::string());
        }
        if (failPhase == 2)
            return -1; // TODO: rethink the return value
        // Line 13
        {
            FilterCoinPH filter(MessageCoinPH2::type, idToss);
            messageList = receive(idProcess, numProcesses/2+1, filter, l0);
            if (_debugLevel >= 10) {
                std::stringstream s;
                s << "P" << idProcess << ",T" << idToss << ": received " << messageList->size() << " PH2 messages\n";
                std::cerr << s.str();
                s.str(std::string());
                for (auto const& m : *messageList) {
                    auto m2 = static_pointer_cast<MessageCoinPH2>(m);
                    s << "P" << idProcess << ",T" << idToss << ": PH2(v=" << m2->v << ")\n";
                    std::cerr << s.str();
                    s.str(std::string());
                }
            }
            if (terminated) {
                if (_debugLevel >= 10) {
                    std::stringstream s;
                    s << "P" << idProcess << ",T" << idToss << ": TERMINATE\n";
                    std::cerr << s.str();
                }
                Channel::freeMessageList(messageList);
                return -1; // TODO: rethink the return value
            }
        }
        // Lines 14-16
        {
            bool received0 = false;
            for (auto const& m : *messageList)
                if ((static_pointer_cast<MessageCoinPH2>(m))->v == 0) {
                    received0 = true;
                    break;
                }
            result = received0 ? 0 : 1;
        }
        Channel::freeMessageList(messageList);
        statistics.countToss(idToss, result);
    }
    else
        result = -1; // TODO: rethink the return value
    // Up to 3*f messages from this toss may not have been used by this process.
    // Discard them, as they are no longer needed and they just make channel simulation slower.
    discardUnusedMessages(idProcess, l0);
    return result;
}

void SimulatedCommonCoin::reset()
{
    throw new std::runtime_error("SimulatedCommonCoin::reset is not implemented.");
}


void SimulatedCommonCoin::terminate()
{
    // TODO: no funciona del todo bien en combinación con RandomConsensus
    std::unique_lock<std::mutex> l0(generalMutex);
    // Marks the end of execution and awakes all blocked processes.
    terminated = true;
    channel.breakWait();
}


bool FailPoint::operator < (const FailPoint& other) const
{
    return failPoint < other.failPoint;
}


bool FailPoint::operator() (const FailPoint& x, const FailPoint& y) const
{
    return x < y;
}


std::ostream& operator << (std::ostream& out, const FailPoint& failPoint)
{
    std::stringstream s;

    s << "FP[pid=" << failPoint.idProcess << ",p=" << failPoint.failPoint << ']';
    out << s.str();
    return out;
}


ArrayCoin::ArrayCoin(unsigned int _numProcesses, unsigned int _maxFailures) :
	Coin(_numProcesses, _maxFailures)
{
    v = new int[numProcesses];
	messages = new int[numProcesses];
	processIndex = new unsigned int[numProcesses];
	receiveIndex = new unsigned int[numProcesses];
	for (unsigned int i = 0; i < numProcesses; i++)
        processIndex[i] = receiveIndex[i] = i;
	failPoints.resize(maxFailures);
}


ArrayCoin::~ArrayCoin()
{
    delete[] v;
    delete[] messages;
    delete[] processIndex;
}


void ArrayCoin::chooseFailPoints()
{
    for (unsigned int i = 0; i < maxFailures; i++) {
        failPoints[i].idProcess = i;
        failPoints[i].failPoint = random() % (3 * numProcesses);
    }
	std::sort(failPoints.begin(), failPoints.end());
}


void ArrayCoin::markFailures()
{
    while (nextFailPoint < maxFailures && failPoints[nextFailPoint].failPoint < currentIteration) {
        v[failPoints[nextFailPoint].idProcess] = Coin::COIN_FAILURE;
        messages[failPoints[nextFailPoint].idProcess] = Coin::COIN_FAILURE;
        nextFailPoint++;
    }
}


void ArrayCoin::doPhase(std::string phaseName)
{
    // broadcast
	for (unsigned int i = 0; i < numProcesses; i++)
		messages[i] = v[i];

    // receive
    shuffle((int*) processIndex, numProcesses);
	for (unsigned int i = 0; i < numProcesses; i++) {
        const unsigned int idProcess = processIndex[i];
        currentIteration++;
        if (_debugLevel >= 5) {
            std::stringstream s;
            s << "P" << idProcess << ": " << phaseName << ", currIt=" << currentIteration << '\n';
            std::cerr << s.str();
        }
        markFailures();
        if (v[idProcess] != Coin::COIN_FAILURE) {
            shuffle((int*) receiveIndex, numProcesses);
            v[idProcess] = 1;
            unsigned int numMessagesReceived = 0;
            for (unsigned int j = 0; j < numProcesses && numMessagesReceived < numProcesses/2+1 && v[idProcess] == 1; j++)
                if (messages[receiveIndex[j]] != Coin::COIN_FAILURE) {
                    numMessagesReceived++;
                    if (messages[receiveIndex[j]] == 0)
                        v[idProcess] = 0;
                }
            if (_debugLevel >= 5) {
                std::stringstream s;
                s << "nMR=" << numMessagesReceived << ", v[i]=" << v[idProcess] << '\n';
                std::cerr << s.str();
            }
        }
        else if (_debugLevel >= 5)
            std::cerr << " -> FAILED\n";
	}
	if (_debugLevel >= 5) {
		std::cerr << phaseName << ": v[i] = ";
		show(v, numProcesses);
	}
}


void ArrayCoin::toss(int* results)
{
	chooseFailPoints();
	nextFailPoint = 0;
	currentIteration = 0;
	if (_debugLevel >= 5) {
        std::stringstream s;
        s << "Fail Points: ";
        for (unsigned int i = 0; i < maxFailures; i++)
            s << failPoints[i] << ';';
        s << '\n';
        std::cerr << s.str() << std::flush;
	}
// Line 1
	for (unsigned int i = 0; i < numProcesses; i++)
		v[i] = individualToss();
	if (_debugLevel >= 5) {
		cout << "START: v[i] = ";
		show(v, numProcesses);
	}
// Lines 2-6
    doPhase("PH0");
// Lines 7-11
    doPhase("PH1");
// Lines 12-16
    doPhase("PH2");
	for (unsigned int i = 0; i < numProcesses; i++)
		results[i] = v[i];
}


