/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#ifndef __COIN_H__
#define __COIN_H__

#include <mutex>
#include <condition_variable>
#include <map>
#include <vector>
#include "Channel.h"

class Coin {
public:
	Coin(unsigned int numProcesses, unsigned int maxFailures);
	virtual ~Coin();
	virtual void toss(int* results) = 0;
	virtual void debug(int level);
	int getDebug() const { return _debugLevel; }
	static const int COIN_FAILURE = -1;
	void show(int* values, unsigned int numValues);
protected:
	virtual int individualToss();
	void shuffle(int* messages, unsigned int numMessages);
	unsigned int numProcesses, maxFailures;
	int _debugLevel;
};

class SimpleCoin : public Coin {
public:
	SimpleCoin(unsigned int numProcesses, unsigned int maxFailures);
	virtual ~SimpleCoin();
	virtual void toss(int* results);
};


struct CommonToss {
    CommonToss(unsigned int numProcesses, unsigned int maxFailures);
    const unsigned int numProcesses, maxFailures; /**< Value of N,F in this toss. */
    unsigned int numParticipants; /**< Number of participant process (actual) in this toss. */
    unsigned int numPendingConsumers; /**< Number of processes which have not consumed this toss yet. */
    int* individualTosses; /**< Individual values in this toss. */
    std::condition_variable cvQuorum; /**< Used to wait until there are enough participants. */
};


class CommonCoin : public Coin {
public:
	CommonCoin(unsigned int numProcesses, unsigned int maxFailures);
	virtual ~CommonCoin();
	virtual void toss(int* results);
	virtual int toss(int idProcess, unsigned int failureType);
	virtual void reset() = 0;
	virtual void terminate() = 0;
protected:
    unsigned int* numTossProcesses; /**< Number of last toss in which each process has participated. */
    std::map<int,CommonToss*> tosses; /**< Tosses which have been done but have not been fully consumed. */
	bool terminated; /**< True if processes have finished executing. */
private:
	std::mutex generalMutex;
};


struct FailPoint {
    unsigned int idProcess;
    unsigned int failPoint;
    bool operator < (const FailPoint& other) const;
    bool operator() (const FailPoint& x, const FailPoint& y) const;
    friend std::ostream& operator << (std::ostream& out, const FailPoint& failPoint);
};


class ArrayCoin : public Coin {
public:
	ArrayCoin(unsigned int numProcesses, unsigned int maxFailures);
	virtual ~ArrayCoin();
	virtual void toss(int* results);
private:
	int* v;
	int* messages;
	unsigned int* processIndex;
	unsigned int* receiveIndex;
	std::vector<FailPoint> failPoints;
	unsigned int currentIteration;
	unsigned int nextFailPoint;
	void chooseFailPoints();
	void markFailures();
	void doPhase(std::string phaseName);
};


class SyntheticCommonCoin : public CommonCoin {
public:
	SyntheticCommonCoin(unsigned int numProcesses, unsigned int maxFailures, double bias);
	virtual ~SyntheticCommonCoin();
	virtual void toss(int* results);
	virtual int toss(int idProcess, unsigned int failureType);
	virtual void reset();
	virtual void terminate();
private:
	std::mutex generalMutex;
	static const int rangeRandom = 1000000;
	int biasRandom;
};


class MessageCoinPH0 : public Message {
public:
    MessageCoinPH0(unsigned int idToss, int v);
    ~MessageCoinPH0();
    static const int type = 0x00001000;
    const unsigned int idToss;
    const int v;
};


class MessageCoinPH1 : public Message {
public:
    MessageCoinPH1(unsigned int idToss, int v);
    ~MessageCoinPH1();
    static const int type = 0x00001001;
    const unsigned int idToss;
    const int v;
};


class MessageCoinPH2 : public Message {
public:
    MessageCoinPH2(unsigned int idToss, int v);
    ~MessageCoinPH2();
    static const int type = 0x00001002;
    const unsigned int idToss;
    const int v;
};


struct CommonCoinStatistics {
    static const int MAX_V = 2;
    CommonCoinStatistics(unsigned int n, unsigned int f);
    unsigned int n, f;
    unsigned numTosses, numCommonTosses;
    unsigned commonValue[MAX_V]; /**< Counter per common values. */
    CommonCoinStatistics& operator += (const CommonCoinStatistics& other);
    friend std::ostream& operator << (std::ostream& out, const CommonCoinStatistics& statistics);
    void outputXml(bool xml) { this->xml = xml; };
    void countToss(unsigned int idToss, int value);
    void countToss(int* results);
    static const int NOT_TOSSED = -1;
    static const int NOT_COMMON = -2;
private:
    bool xml;
    std::vector<int> tossResults;
    static const int INC_SIZE = 1000;
};


class SimulatedCommonCoin : public CommonCoin {
public:
	SimulatedCommonCoin(unsigned int numProcesses, unsigned int maxFailures, Channel& channel);
	virtual ~SimulatedCommonCoin();
	virtual int toss(int idProcess, unsigned int failureType);
	virtual void reset();
	virtual void terminate();
	CommonCoinStatistics statistics;
private:
    class FilterCoinPH;
    class FilterCoinPreviousTosses;
    Channel& channel;
	std::mutex generalMutex;
    MessageList* receive(unsigned int idProcess, unsigned int howMany, MessageFilter& filter, std::unique_lock<std::mutex>& coinLock);
    void discardUnusedMessages(unsigned int idProcess, std::unique_lock<std::mutex>& coinLock);
};

extern std::string coinProb0;

#endif
