/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#include <iostream>
#include <iomanip>      // std::setprecission
#include <sstream>
#include <memory>       // std::static_pointer_cast
#include <thread>       // std::this_thread::sleep_for
#include "DRConsensus.h"
#include "Arguments.h"      // Failure types


DRConsensusStatistics::DRConsensusStatistics(unsigned int _n) :
    DRConsensusStatistics(_n, 0)
{
}

DRConsensusStatistics::DRConsensusStatistics(unsigned int _n, unsigned int _f) :
    numExecutions(0),
    numCorrectExecutions(0),
    n(_n),
    f(_f),
    xml(false)
{
    for (unsigned int i = 0; i < MAX_V; i++)
        proposed[i] = 0;
    for (unsigned int i = 0; i < MAX_V; i++)
        decided[i] = 0;
}

DRConsensusStatistics& DRConsensusStatistics::operator += (const DRConsensusStatistics& other)
{
    numExecutions += other.numExecutions;
    numCorrectExecutions += other.numCorrectExecutions;
    numRounds += other.numRounds;
    for (unsigned int i = 0; i < MAX_V; i++)
        proposed[i] += other.proposed[i];
    for (unsigned int i = 0; i < MAX_V; i++)
        decided[i] += other.decided[i];
    return *this;
}


std::ostream& operator << (std::ostream& out, const DRConsensusStatistics& statistics)
{
    std::stringstream s;
    unsigned int total_proposed, total_decided;

//    std::setprecision(4);
    if (statistics.xml) {
        s << "<drandom_consensus>";
        s << "<num_processes>" << statistics.n << "</num_processes><max_failures>" << statistics.f << "</max_failures>";
        s << "<num_executions>" << statistics.numExecutions << "</num_executions></num_correct_executions>" << statistics.numCorrectExecutions << "</num_correct_executions>";
    }
    else {
        s << "n=" << statistics.n << ", f=" << statistics.f << '\n';
        s << "ex=" << statistics.numExecutions << ", cex=" << statistics.numCorrectExecutions << '\n';
    }

    if (statistics.xml)
        s << "<num_rounds_avg>" << statistics.numRounds.average() << "</num_rounds_avg>";
    else
        s << "Num rounds: " << statistics.numRounds << '\n';

    if (statistics.xml)
        s << "<proposals>";
    total_proposed = 0;
    for (unsigned int i = 0; i < statistics.MAX_V; i++)
        total_proposed += statistics.proposed[i];
    for (unsigned int i = 0; i < statistics.MAX_V; i++)
        if (statistics.xml)
            s << "<case><v>" << i << "</v><num>" << statistics.proposed[i] << "</num></case>";
        else
            s << "proposed[" << i << "] = " << statistics.proposed[i] << " (" << (double) statistics.proposed[i] / total_proposed * 100.0 << "%)\n";
    if (statistics.xml)
        s << "</proposals>";

    if (statistics.xml)
        s << "<decissions>";
    total_decided = 0;
    for (unsigned int i = 0; i < statistics.MAX_V; i++)
        total_decided += statistics.decided[i];
    for (unsigned int i = 0; i < statistics.MAX_V; i++)
        if (statistics.xml)
            s << "<case><v>" << i << "</v><num>" << statistics.decided[i] << "</num></case>";
        else
            s << "decided[" << i << "] = " << statistics.decided[i] << " (" << (double) statistics.decided[i] / total_decided * 100.0 << "%)\n";
    if (statistics.xml)
        s << "</decissions>";

    if (statistics.xml)
        s << "</drandom_consensus>\n";
    out << s.str();
    return out;
}


DRConsensus::FilterPhase1::FilterPhase1(unsigned int _r) :
    r(_r)
{
}


bool DRConsensus::FilterPhase1::operator()(std::shared_ptr<const Message> _mensaje)
{
    if (_mensaje->type == MessagePROP::type) {
        std::shared_ptr<const MessagePROP> mensaje = std::static_pointer_cast<const MessagePROP>(_mensaje);
        return r==mensaje->r;
    }
    else if (_mensaje->type == MessageCHECK::type) {
        std::shared_ptr<const MessageCHECK> mensaje = std::static_pointer_cast<const MessageCHECK>(_mensaje);
        return r==mensaje->r;
    }
    else if (_mensaje->type == MessageDECIDE::type)
        return true;
    else
        return false;
}


DRConsensus::DRConsensus(unsigned int _numProcesses, int _idProcess, Channel& _channel, CommonCoin& _coin, SyntheticFailureDetector& _failureDetector, unsigned int _failureType) :
	RandomConsensus(_numProcesses, _idProcess, _channel, _coin, _failureType),
	statistics(_numProcesses),
	failureDetector(_failureDetector)
{
}


DRConsensus::~DRConsensus()
{
}


int DRConsensus::propose(int v)
{
    int d;

    statistics.numExecutions++;
    if (failureType == ProgramArguments::FAILURE_START)
        return -1;
    decide = 0;
	// Line 1
	r = 0;
	// Line 2
	est = v;
	// Line 3
	// In this implementation, tasks 1 and 2 are not launched as separate threads.
	// Task 1 will check for DECIDE messages whenever a message is received.
	d = task1();
	if (d != -1) {
        statistics.numCorrectExecutions++;
        statistics.numRounds.tmpValue = r;
        statistics.numRounds.collect();
        statistics.proposed[v]++;
        statistics.decided[d]++;
    }
    coin.terminate();
	return d;
}


int DRConsensus::task1()
{
    MessageList* messageList;
    std::shared_ptr<Message> mCHECKWait0; // This is used to keep a CHECK message that was received in line 7-c, but that has to be processed in line 11

    // Line 4
    do {
        // Line 5
        r++;
        if (_debugLevel >= 5) {
            std::stringstream s;
            s << std::string(idProcess*2, ' ') << "P" << idProcess << ": round " << r << " (est=" << est << ")\n";
            std::cerr << s.str();
            s.str(std::string());
        }
        // Line 6
        ld = failureDetector.amLeader(idProcess);
        if (ld)
            channel.broadcast(idProcess, std::make_shared<MessagePROP>(r, est));
        if (_debugLevel >= 3) {
            std::stringstream s;
            if (ld)
                s << std::string(idProcess*2, ' ') << "P" << idProcess << ": broadcast PROP(r=" << r << ",est=" << est << ")\n";
            else
                s << std::string(idProcess*2, ' ') << "P" << idProcess << ": not leader\n";
            std::cerr << s.str();
        }
        // Lines 7-a to 7-c
        {
            if (_debugLevel >= 20) {
                std::stringstream s;
                s << std::string(idProcess*2, ' ') << "P" << idProcess << ": WAIT L7\n";
                std::cerr << s.str();
            }
            messageList = receivePhase1();
            if (_debugLevel >= 10) {
                std::stringstream s;
                s << std::string(idProcess*2, ' ') << "P" << idProcess << ": received " << messageList->size() << " messages\n";
                std::cerr << s.str();
                s.str(std::string());
                for (auto const& m : *messageList) {
                    switch(m->type) {
                    case MessagePROP::type:
                        {
                            auto mprop = std::static_pointer_cast<MessagePROP>(m);
                            s << std::string(idProcess*2, ' ') << "P" << idProcess << ": PROP(r=" << mprop->r << ",est=" << mprop->est << ")\n";
                        }
                        break;
                    case MessageCHECK::type:
                        {
                            auto mcheck = std::static_pointer_cast<MessageCHECK>(m);
                            s << std::string(idProcess*2, ' ') << "P" << idProcess << ": CHECK(r=" << mcheck->r << ",common_est=" << mcheck->common_est << ")\n";
                        }
                        break;
                    case MessageDECIDE::type:
                        {
                            auto mdecide = std::static_pointer_cast<MessageDECIDE>(m);
                            s << std::string(idProcess*2, ' ') << "P" << idProcess << ": DECIDE(v=" << mdecide->v << ")\n";
                        }
                        break;
                    default:
                        s << std::string(idProcess*2, ' ') << "P" << idProcess << ": UNKNOWN MESSAGE TYPE. THIS IS AN ERROR.\n";
                    }
                    std::cerr << s.str();
                    s.str(std::string());
                }
            }
            for (auto const& m : *messageList) {
                if (m->type == MessageDECIDE::type) {
                    if (_debugLevel >= 10) {
                        std::stringstream s;
                        s << std::string(idProcess*2, ' ') << "P" << idProcess << ": received DECIDE(" << decide->v << ")\n";
                        std::cerr << s.str();
                    }
                    Channel::freeMessageList(messageList);
                    return std::static_pointer_cast<MessageDECIDE>(m)->v;
                }
            }
        }
        // Line 8
        // If any messages were received, they were PROP and/or CHECK. DECIDE messages would have been processed above.
        // TODO: JMR what if they are mixed?
        mCHECKWait0 = 0;
        if (messageList->size() > 0) {
            unsigned int num0 = 0;
            for (auto const& m : *messageList) {
                switch(m->type) {
                case MessagePROP::type:
                    {
                        auto mprop = std::static_pointer_cast<MessagePROP>(m);
                        if (mprop->est == 0)
                            num0++;
                    }
                    break;
                case MessageCHECK::type:
                    {
                        auto mcheck = std::static_pointer_cast<MessageCHECK>(m);
                        if (mcheck->common_est == 0)
                            num0++;
                        mCHECKWait0 = m;
                    }
                    break;
                default:
                    throw new std::runtime_error("Unknown message type");
                }
            }
            if (messageList->size() - num0 >= num0)
                common_est = 1;
            else
                common_est = 0;
        }
        else
            common_est = est;
        Channel::freeMessageList(messageList);
        // Line 9
        channel.broadcast(idProcess, std::make_shared<MessageCHECK>(r, common_est));
        if (_debugLevel >= 3) {
            std::stringstream s;
            s << std::string(idProcess*2, ' ') << "P" << idProcess << ": broadcast CHECK(r=" << r << ",common_est=" << common_est << ")\n";
            std::cerr << s.str();
            s.str(std::string());
        }
        // Line 10
        {
            FilterCHECK filter(r, &decide);
            if (_debugLevel >= 20) {
                std::stringstream s;
                s << std::string(idProcess*2, ' ') << "P" << idProcess << ": WAIT L10\n";
                std::cerr << s.str();
            }
            // If a CHECK message was already received in line 7-c, one less message has to be received here.
            messageList = receive(numProcesses/2 + (mCHECKWait0 == 0 ? 1 : 0), filter);
            if (_debugLevel >= 10) {
                std::stringstream s;
                s << std::string(idProcess*2, ' ') << "P" << idProcess << ": received " << messageList->size() << " CHECK messages\n";
                std::cerr << s.str();
                s.str(std::string());
                for (auto const& m : *messageList) {
                    auto m2 = std::static_pointer_cast<MessageCHECK>(m);
                    s << std::string(idProcess*2, ' ') << "P" << idProcess << ": CHECK(r=" << m2->r << ",common_est=" << m2->common_est << ")\n";
                    std::cerr << s.str();
                    s.str(std::string());
                }
            }
            if (decide != 0) {
                if (_debugLevel >= 10) {
                    std::stringstream s;
                    s << std::string(idProcess*2, ' ') << "P" << idProcess << ": received DECIDE(" << decide->v << ")\n";
                    std::cerr << s.str();
                }
                Channel::freeMessageList(messageList);
                return decide->v;
            }
            // If a CHECK message was already received in line 7-c, include it here, so that it will be processed by lines 11-13
            if (mCHECKWait0 != 0)
                messageList->push_front(mCHECKWait0);
        }
        // Lines 11-13
        {
            same = true;
            for (auto const& m : *messageList)
                if ((std::static_pointer_cast<const MessageCHECK>(m))->common_est != common_est) {
                    same = false;
                    break;
                }
        }
        Channel::freeMessageList(messageList);
        // Line 14
        channel.broadcast(idProcess, std::make_shared<MessageAGREE>(r, common_est, same));
        if (_debugLevel >= 3) {
            std::stringstream s;
            s << std::string(idProcess*2, ' ') << "P" << idProcess << ": broadcast AGREE(r=" << r << ",common_est=" << common_est << ",same=" << same << ")\n";
            std::cerr << s.str();
            s.str(std::string());
        }
        // Line 15
        {
            FilterAGREE filter(r, &decide);
            if (_debugLevel >= 20) {
                std::stringstream s;
                s << std::string(idProcess*2, ' ') << "P" << idProcess << ": WAIT L15\n";
                std::cerr << s.str();
            }
            messageList = receive(numProcesses/2+1, filter);
            if (decide != 0) {
                if (_debugLevel >= 10) {
                    std::stringstream s;
                    s << std::string(idProcess*2, ' ') << "P" << idProcess << ": received DECIDE(" << decide->v << ")\n";
                    std::cerr << s.str();
                }
                Channel::freeMessageList(messageList);
                return decide->v;
            }
        }
        // Lines 16-18
        {
            for (auto const& m : *messageList) {
                std::shared_ptr<const MessageAGREE> m2 = std::static_pointer_cast<const MessageAGREE>(m);
                if (m2->same) {
                    est = m2->common_est;
                    break;
                }
            }
        }
        // Lines 19-21
        {
            bool unanimity = true;
            for (auto const& m : *messageList) {
                std::shared_ptr<const MessageAGREE> m2 = std::static_pointer_cast<const MessageAGREE>(m);
                if (! m2->same) {
                    unanimity = false;
                    break;
                }
            }
            if (unanimity) {
                std::shared_ptr<const MessageAGREE> m = std::static_pointer_cast<const MessageAGREE>(messageList->front());
                channel.broadcast(idProcess, std::make_shared<MessageDECIDE>(m->common_est));
                if (_debugLevel >= 3) {
                    std::stringstream s;
                    s << std::string(idProcess*2, ' ') << "P" << idProcess << ": broadcast DECIDE(common_est=" << m->common_est << ")\n";
                    std::cerr << s.str();
                    s.str(std::string());
                }
                int v = m->common_est;
                Channel::freeMessageList(messageList);
                return v;
            }
        }
        // Line 22
        if (_debugLevel >= 20) {
            std::stringstream s;
            s << std::string(idProcess*2, ' ') << "P" << idProcess << ": coin.toss()\n";
            std::cerr << s.str();
        }
        c = coin.toss(idProcess, failureType);
        // Lines 23-25
        {
            bool unanimity = true;
            for (auto const& m : *messageList) {
                std::shared_ptr<const MessageAGREE> m2 = std::static_pointer_cast<const MessageAGREE>(m);
                if (m2->same) {
                    unanimity = false;
                    break;
                }
            }
            if (unanimity)
                est = c;
        }
        Channel::freeMessageList(messageList);
        failureDetector.advanceRound(idProcess);
    } while (1);
}


MessageList* DRConsensus::receivePhase1()
{
    MessageList* messageListPROP;
    MessageList* messageListCHECK;
    FilterPhase1 filter(r);
    bool finish;

    messageListPROP = new MessageList;
    messageListCHECK = new MessageList;
    finish = false;
    do {
        std::shared_ptr<Message> message = channel.receive(idProcess, filter, false);
        if (message != 0) {
            if (_debugLevel >= 12) {
                std::stringstream s;
                s << std::string(idProcess*2, ' ') << "P" << idProcess << ": received message of type " << message->type << '\n';
                std::cerr << s.str();
            }
            switch(message->type) {
            case MessageDECIDE::type:
                decide = std::static_pointer_cast<MessageDECIDE>(message);
                break;
            case MessagePROP::type:
                messageListPROP->push_back(message);
                break;
            case MessageCHECK::type:
                messageListCHECK->push_back(message);
                break;
            }
        }
        if (ld != failureDetector.amLeader(idProcess) ||
            (ld && messageListPROP->size() >= failureDetector.numLeaders(idProcess)) ||
            messageListCHECK->size() > 0 ||
            decide != 0)
            finish = true;
        if (! finish)
            std::this_thread::sleep_for(std::chrono::milliseconds(random() % 20));
    } while (! finish);
    if (_debugLevel >= 5) {
        // Does not test the failure detector because in the unstable state may provide random values.
        if (messageListPROP->size() < failureDetector.numLeaders(idProcess) &&
            messageListCHECK->size() == 0 &&
            decide == 0) {
                std::stringstream s;
                s << std::string(idProcess*2, ' ') << "P" << idProcess << ": leader status changed\n";
                std::cerr << s.str();
            }
    }
    if (messageListCHECK->size() > 0)
        return messageListCHECK;
    else
        return messageListPROP;
}
