/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#include "FailureDetector.h"
#include <stdexcept>        // runtime_error

SyntheticFailureDetector::SyntheticFailureDetector(unsigned int _numProcesses, unsigned int _maxFailures, unsigned int _numUnstableRounds) :
    numProcesses(_numProcesses),
    maxFailures(_maxFailures),
    numUnstableRounds(_numUnstableRounds)
{
    round = new unsigned int[numProcesses];
    for (unsigned int i = 0; i < numProcesses; i++)
        round[i] = 1;
    // Initially, chooses about 20% processes as eventual leaders. Adjustments will be made if later some process indicates it may fail.
    leader = new char[numProcesses];
    for (unsigned int i = 0; i < numProcesses; i++)
        leader[i] = (random() % 5) == 0 ? LEADER : NOT_LEADER;
    numLeaderProcesses = 0;
    for (unsigned int i = 0; i < numProcesses; i++)
        if (leader[i] == LEADER)
            numLeaderProcesses++;
    // Ensures at least one process is marked as leader
    if (numLeaderProcesses == 0) {
        leader[0] = LEADER;
        numLeaderProcesses = 1;
    }
}


SyntheticFailureDetector::~SyntheticFailureDetector()
{
    delete[] leader;
    delete[] round;
}


void SyntheticFailureDetector::mayFail(int idProcess)
{
    // If this process was not going to be a leader, just mark it as a possible failed process.
    if (leader[idProcess] == NOT_LEADER)
        leader[idProcess] = MAY_FAIL;
    // However, if it was going to be a leader, promote some other process to eventual leader.
    else {
        leader[idProcess] = MAY_FAIL;
        numLeaderProcesses--;
        unsigned int n = numProcesses;
        unsigned int i = random() % numProcesses;
        bool promoted = false;
        do {
            if (leader[i] == NOT_LEADER) {
                leader[i] = LEADER;
                numLeaderProcesses++;
                promoted = true;
            }
            else {
                i = (i+1) % numProcesses;
                n--;
            }
        } while (n > 0 && ! promoted);
        if (numLeaderProcesses == 0)
            throw new std::runtime_error("All processes may fail, which means no one can be an eventual leader in the failure detector");
    }
}


bool SyntheticFailureDetector::amLeader(int idProcess) const
{
    if (round[idProcess] <= numUnstableRounds)
//        return (random() % 2) == 0;
        return true;
    else
        return leader[idProcess] == LEADER;
}


unsigned int SyntheticFailureDetector::numLeaders(int idProcess) const
{
    if (round[idProcess] <= numUnstableRounds)
        return random() % numProcesses;
//        return numProcesses/2-1;
    else
        return numLeaderProcesses;
}


void SyntheticFailureDetector::advanceRound(int idProcess)
{
    round[idProcess]++;
}
