/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#ifndef __FAILUREDETECTOR_H__
#define __FAILUREDETECTOR_H__


/**
 * Represents a synthetic Omega Anonymous Failure Detector.
 * The implementation is not based on any specific distributed algorithm,
 * and it just provides synthesized results.
 *
 * The failure detector has two periods of behaviour.
 * For the first R rounds it behaves unstably, returning inconsistent hints to all processes.
 * At round R+1 it enters into the stable period, during which consistent fixed hints are given.
 * Note that it is possible that some processes are in the unstable period, while other processes have
 * already reached the stable period. In such case, consistent information will be given only to
 * processes in the stable period.
 *
 * Note that process identifiers are used in several methods for simulation purposes, but the actual
 * behaviour mimics an anonymous system.
 *
 * Concurrent access to objects of this class is possible, except in the mayFail() method. That should
 * not be a problem, as it is easy to invoke that method non-concurrently, during the initialization stage of the simulation.
 */
class SyntheticFailureDetector {
public:
    /**
     * Create a synthetic Omega Anonymous Failure Detector.
     * @param numProcesses Number of participating processes.
     * @param maxFailures Maximum number of processes which may fail during execution.
     * @param numUnstableRounds Number of rounds (R) of the unstable period. Stability will start at round R+1.
     */
    SyntheticFailureDetector(unsigned int numProcesses, unsigned int maxFailures, unsigned int numUnstableRounds);
    /**
     * Destroy a synthetic Omega Anonymous Failure Detector.
     */
    virtual ~SyntheticFailureDetector();
    /**
     * Indicate that a given process may fail during execution, which means that
     * it would not be a suitable leader in the stable period.
     * These indications must be given before the consensus algorithm starts.
     * Otherwise, the failure detector may not work correctly.
     * @param idProcess Identifier of the process which may fail.
     */
    void mayFail(int idProcess);
    /**
     * Return a hint whether the invoking process is a leader.
     * Note that this information will be inconsistent during the unstable period.
     * This method is not reentrant.
     * @param idProcess Identifier of the invoking process.
     * @return True if the invoking process is a leader.
     */
    bool amLeader(int idProcess) const;
    /**
     * Return an estimation of the number of leaders in the system.
     * Note that this information will be inconsistent during the unstable period.
     * @param idProcess Identifier of the invoking process.
     * @return Estimated number of leaders in the system.
     */
    unsigned int numLeaders(int idProcess) const;
    /**
     * Indicate that the invoking process is advancing to the next round in the consensus algorithm.
     * This method must be used at the end of each round of the consensus algorithm, so that the failure
     * detector knows when the unstable period finishes.
     * @param idProcess Identifier of the invoking process.
     */
    void advanceRound(int idProcess);
private:
    unsigned int numProcesses;
    unsigned int maxFailures;
    unsigned int numUnstableRounds;
    unsigned int* round;
    char* leader;
    static const char LEADER = 1;
    static const char NOT_LEADER = 2;
    static const char MAY_FAIL = 3;
    unsigned int numLeaderProcesses;
};

#endif // __FAILUREDETECTOR_H__
