/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#ifndef __OBSERVATIONS_H__
#define __OBSERVATIONS_H__

#include <list>
#include <ostream>
#include <sstream>
#include <math.h>


template <class TYPE>
class ObservedVariable
{
public:
    ObservedVariable() : ObservedVariable(0) {};
    ObservedVariable(TYPE v) : tmpValue(v), currentSum(0) {}
    virtual ~ObservedVariable() {};
    ObservedVariable& operator += (const ObservedVariable& other);
    template <class T2> friend std::ostream& operator << (std::ostream& out, const ObservedVariable<T2>& variable);
    void collect();
    inline TYPE sum() const { return currentSum; };
    inline TYPE min() const { return currentMin; };
    inline TYPE max() const { return currentMax; };
    double average() const;
    double stdDev() const;
    TYPE tmpValue;
    std::list<TYPE> observations;
private:
    TYPE currentSum;
    TYPE currentMin;
    TYPE currentMax;
};


template <class TYPE>
ObservedVariable<TYPE>& ObservedVariable<TYPE>::operator += (const ObservedVariable<TYPE>& other)
{
    if (other.observations.size() > 0) {
        if (this->observations.size() == 0) {
            this->currentMin = other.currentMin;
            this->currentMax = other.currentMax;
        }
        else {
            this->currentMin = std::min<TYPE>(this->currentMin, other.currentMin);
            this->currentMax = std::max<TYPE>(this->currentMax, other.currentMax);
        }
    }
    this->tmpValue += other.tmpValue;
    this->observations.insert(this->observations.end(), other.observations.begin(), other.observations.end());
    this->currentSum += other.currentSum;
    return *this;
}


template <class TYPE>
std::ostream& operator << (std::ostream& out, const ObservedVariable<TYPE>& variable)
{
    std::stringstream s;

    s << "tmpValue: " << variable.tmpValue << ", sum: " << variable.sum() << ", avg: " << variable.average() << ", stdDev: " << variable.stdDev() << " (" << variable.stdDev() / variable.average() * 100.0 << "%), min: " << variable.currentMin << ", max: " << variable.currentMax << ", observations: {";
    for (auto const& o : variable.observations)
        s << ' ' << o;
    s << " }";
    out << s.str();
    return out;
}


template <class TYPE>
void ObservedVariable<TYPE>::collect()
{
    currentSum += tmpValue;
    if (observations.size() == 0)
        currentMin = currentMax = tmpValue;
    else {
        currentMin = std::min<TYPE>(currentMin, tmpValue);
        currentMax = std::max<TYPE>(currentMax, tmpValue);
    }
    observations.push_back(tmpValue);
    tmpValue = 0;
}


template <class TYPE>
double ObservedVariable<TYPE>::average() const
{
    return 1.0 * currentSum / observations.size();
}


template <class TYPE>
double ObservedVariable<TYPE>::stdDev() const
{
    double m;
    double sd;

    m = average();
    sd = 0;
    for (auto const& o : observations)
        sd += (o - m) * (o - m);
    sd /= observations.size();
    return sqrt(sd);
}


#endif // __OBSERVATIONS_H__
