/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#include <iostream>
#include <iomanip>      // std::setprecission
#include <sstream>
#include <memory>       // std::static_pointer_cast
#include "RandomConsensus.h"
#include "Arguments.h"      // Failure types

RandomConsensusStatistics::RandomConsensusStatistics(unsigned int _n) :
    RandomConsensusStatistics(_n, 0)
{
}

RandomConsensusStatistics::RandomConsensusStatistics(unsigned int _n, unsigned int _f) :
    numExecutions(0),
    numCorrectExecutions(0),
    n(_n),
    f(_f),
    xml(false)
{
    for (unsigned int i = 0; i < MAX_V; i++)
        proposed[i] = 0;
    for (unsigned int i = 0; i < MAX_V; i++)
        decided[i] = 0;
}

RandomConsensusStatistics& RandomConsensusStatistics::operator += (const RandomConsensusStatistics& other)
{
    numExecutions += other.numExecutions;
    numCorrectExecutions += other.numCorrectExecutions;
    numRounds += other.numRounds;
    for (unsigned int i = 0; i < MAX_V; i++)
        proposed[i] += other.proposed[i];
    for (unsigned int i = 0; i < MAX_V; i++)
        decided[i] += other.decided[i];
    return *this;
}


std::ostream& operator << (std::ostream& out, const RandomConsensusStatistics& statistics)
{
    std::stringstream s;
    unsigned int total_proposed, total_decided;

//    std::setprecision(4);
    if (statistics.xml) {
        s << "<random_consensus>";
        s << "<num_processes>" << statistics.n << "</num_processes><max_failures>" << statistics.f << "</max_failures>";
        s << "<num_executions>" << statistics.numExecutions << "</num_executions></num_correct_executions>" << statistics.numCorrectExecutions << "</num_correct_executions>";
    }
    else {
        s << "n=" << statistics.n << ", f=" << statistics.f << '\n';
        s << "ex=" << statistics.numExecutions << ", cex=" << statistics.numCorrectExecutions << '\n';
    }

    if (statistics.xml)
        s << "<num_rounds_avg>" << statistics.numRounds.average() << "</num_rounds_avg>";
    else
        s << "Num rounds: " << statistics.numRounds << '\n';

    if (statistics.xml)
        s << "<proposals>";
    total_proposed = 0;
    for (unsigned int i = 0; i < statistics.MAX_V; i++)
        total_proposed += statistics.proposed[i];
    for (unsigned int i = 0; i < statistics.MAX_V; i++)
        if (statistics.xml)
            s << "<case><v>" << i << "</v><num>" << statistics.proposed[i] << "</num></case>";
        else
            s << "proposed[" << i << "] = " << statistics.proposed[i] << " (" << (double) statistics.proposed[i] / total_proposed * 100.0 << "%)\n";
    if (statistics.xml)
        s << "</proposals>";

    if (statistics.xml)
        s << "<decissions>";
    total_decided = 0;
    for (unsigned int i = 0; i < statistics.MAX_V; i++)
        total_decided += statistics.decided[i];
    for (unsigned int i = 0; i < statistics.MAX_V; i++)
        if (statistics.xml)
            s << "<case><v>" << i << "</v><num>" << statistics.decided[i] << "</num></case>";
        else
            s << "decided[" << i << "] = " << statistics.decided[i] << " (" << (double) statistics.decided[i] / total_decided * 100.0 << "%)\n";
    if (statistics.xml)
        s << "</decissions>";

    if (statistics.xml)
        s << "</random_consensus>\n";
    out << s.str();
    return out;
}


RandomConsensus::FilterPROP::FilterPROP(unsigned int _r, std::shared_ptr<const MessageDECIDE>* _decide) :
    r(_r),
    decide(_decide)
{
}


bool RandomConsensus::FilterPROP::operator()(std::shared_ptr<const Message> _mensaje)
{
    if (_mensaje->type == MessagePROP::type) {
        std::shared_ptr<const MessagePROP> mensaje = std::static_pointer_cast<const MessagePROP>(_mensaje);
        return r==mensaje->r;
    }
    else if (_mensaje->type == MessageDECIDE::type) {
        std::shared_ptr<const MessageDECIDE> mensaje = std::static_pointer_cast<const MessageDECIDE>(_mensaje);
        *decide = mensaje;
        return true;
    }
    else
        return false;
}


RandomConsensus::FilterCHECK::FilterCHECK(unsigned int _r, std::shared_ptr<const MessageDECIDE>* _decide) :
    r(_r),
    decide(_decide)
{
}


bool RandomConsensus::FilterCHECK::operator()(std::shared_ptr<const Message> _mensaje)
{
    if (_mensaje->type == MessageCHECK::type) {
        std::shared_ptr<const MessageCHECK> mensaje = std::static_pointer_cast<const MessageCHECK>(_mensaje);
        return r==mensaje->r;
    }
    else if (_mensaje->type == MessageDECIDE::type) {
        std::shared_ptr<const MessageDECIDE> mensaje = std::static_pointer_cast<const MessageDECIDE>(_mensaje);
        *decide = mensaje;
        return true;
    }
    else
        return false;
}


RandomConsensus::FilterAGREE::FilterAGREE(unsigned int _r, std::shared_ptr<const MessageDECIDE>* _decide) :
    r(_r),
    decide(_decide)
{
}


bool RandomConsensus::FilterAGREE::operator()(std::shared_ptr<const Message> _mensaje)
{
    if (_mensaje->type == MessageAGREE::type) {
        std::shared_ptr<const MessageAGREE> mensaje = std::static_pointer_cast<const MessageAGREE>(_mensaje);
        return r==mensaje->r;
    }
    else if (_mensaje->type == MessageDECIDE::type) {
        std::shared_ptr<const MessageDECIDE> mensaje = std::static_pointer_cast<const MessageDECIDE>(_mensaje);
        *decide = mensaje;
        return true;
    }
    else
        return false;
}


RandomConsensus::RandomConsensus(unsigned int _numProcesses, int _idProcess, Channel& _channel, CommonCoin& _coin, unsigned int _failureType) :
	Consensus(_numProcesses, _idProcess),
	statistics(_numProcesses),
	channel(_channel),
	coin(_coin),
	failureType(_failureType)
{
}


RandomConsensus::~RandomConsensus()
{
}


int RandomConsensus::propose(int v)
{
    int d;

    statistics.numExecutions++;
    if (failureType == ProgramArguments::FAILURE_START)
        return -1;
    decide = 0;
	// Line 1
	r = 0;
	// Line 2
	est = v;
	// Line 3
	// In this implementation, tasks 1 and 2 are not launched as separate threads.
	// Task 1 will check for DECIDE messages whenever a message is received.
	d = task1();
	if (d != -1) {
        statistics.numCorrectExecutions++;
        statistics.numRounds.tmpValue = r;
        statistics.numRounds.collect();
        statistics.proposed[v]++;
        statistics.decided[d]++;
    }
    coin.terminate();
	return d;
}


int RandomConsensus::task1()
{
    MessageList* messageList;

    // Line 4
    do {
        // Line 5
        r++;
        if (_debugLevel >= 5) {
            std::stringstream s;
            s << "P" << idProcess << ": round " << r << '\n';
            std::cerr << s.str();
            s.str(std::string());
        }
        // Line 6
        channel.broadcast(idProcess, std::make_shared<MessagePROP>(r, est));
        if (_debugLevel >= 3) {
            std::stringstream s;
            s << "P" << idProcess << ": broadcast PROP(r=" << r << ",est=" << est << ")\n";
            std::cerr << s.str();
            s.str(std::string());
        }
        // Line 7
        {
            FilterPROP filter(r, &decide);
            if (_debugLevel >= 20) {
                std::stringstream s;
                s << "P" << idProcess << ": WAIT L7\n";
                std::cerr << s.str();
            }
            messageList = receive(numProcesses/2+1, filter);
            if (_debugLevel >= 10) {
                std::stringstream s;
                s << "P" << idProcess << ": received " << messageList->size() << " PROP messages\n";
                std::cerr << s.str();
                s.str(std::string());
                for (auto const& m : *messageList) {
                    auto m2 = std::static_pointer_cast<MessagePROP>(m);
                    s << "P" << idProcess << ": PROP(r=" << m2->r << ",est=" << m2->est << ")\n";
                    std::cerr << s.str();
                    s.str(std::string());
                }
            }
            if (decide != 0) {
                if (_debugLevel >= 10) {
                    std::stringstream s;
                    s << "P" << idProcess << ": received DECIDE(" << decide->v << ")\n";
                    std::cerr << s.str();
                }
                Channel::freeMessageList(messageList);
                return decide->v;
            }
        }
        // Line 8
        {
            unsigned int num0 = 0;
            for (auto const& m : *messageList)
                if ((std::static_pointer_cast<MessagePROP>(m))->est == 0)
                    num0++;
            if (messageList->size() - num0 >= num0)
                common_est = 1;
            else
                common_est = 0;
        }
        Channel::freeMessageList(messageList);
        // Line 9
        channel.broadcast(idProcess, std::make_shared<MessageCHECK>(r, common_est));
        if (_debugLevel >= 3) {
            std::stringstream s;
            s << "P" << idProcess << ": broadcast CHECK(r=" << r << ",common_est=" << common_est << ")\n";
            std::cerr << s.str();
            s.str(std::string());
        }
        // Line 10
        {
            FilterCHECK filter(r, &decide);
            if (_debugLevel >= 20) {
                std::stringstream s;
                s << "P" << idProcess << ": WAIT L10\n";
                std::cerr << s.str();
            }
            messageList = receive(numProcesses/2+1, filter);
            if (_debugLevel >= 10) {
                std::stringstream s;
                s << "P" << idProcess << ": received " << messageList->size() << " CHECK messages\n";
                std::cerr << s.str();
                s.str(std::string());
                for (auto const& m : *messageList) {
                    auto m2 = std::static_pointer_cast<MessageCHECK>(m);
                    s << "P" << idProcess << ": CHECK(r=" << m2->r << ",common_est=" << m2->common_est << ")\n";
                    std::cerr << s.str();
                    s.str(std::string());
                }
            }
            if (decide != 0) {
                if (_debugLevel >= 10) {
                    std::stringstream s;
                    s << "P" << idProcess << ": received DECIDE(" << decide->v << ")\n";
                    std::cerr << s.str();
                }
                Channel::freeMessageList(messageList);
                return decide->v;
            }
        }
        // Lines 11-13
        {
            same = true;
            for (auto const& m : *messageList)
                if ((std::static_pointer_cast<const MessageCHECK>(m))->common_est != common_est) {
                    same = false;
                    break;
                }
        }
        Channel::freeMessageList(messageList);
        // Line 14
        channel.broadcast(idProcess, std::make_shared<MessageAGREE>(r, common_est, same));
        if (_debugLevel >= 3) {
            std::stringstream s;
            s << "P" << idProcess << ": broadcast AGREE(r=" << r << ",common_est=" << common_est << ",same=" << same << ")\n";
            std::cerr << s.str();
            s.str(std::string());
        }
        // Line 15
        {
            FilterAGREE filter(r, &decide);
            if (_debugLevel >= 20) {
                std::stringstream s;
                s << "P" << idProcess << ": WAIT L15\n";
                std::cerr << s.str();
            }
            messageList = receive(numProcesses/2+1, filter);
            if (decide != 0) {
                if (_debugLevel >= 10) {
                    std::stringstream s;
                    s << "P" << idProcess << ": received DECIDE(" << decide->v << ")\n";
                    std::cerr << s.str();
                }
                Channel::freeMessageList(messageList);
                return decide->v;
            }
        }
        // Lines 16-18
        {
            for (auto const& m : *messageList) {
                std::shared_ptr<const MessageAGREE> m2 = std::static_pointer_cast<const MessageAGREE>(m);
                if (m2->same) {
                    est = m2->common_est;
                    break;
                }
            }
        }
        // Lines 19-21
        {
            bool unanimity = true;
            for (auto const& m : *messageList) {
                std::shared_ptr<const MessageAGREE> m2 = std::static_pointer_cast<const MessageAGREE>(m);
                if (! m2->same) {
                    unanimity = false;
                    break;
                }
            }
            if (unanimity) {
                std::shared_ptr<const MessageAGREE> m = std::static_pointer_cast<const MessageAGREE>(messageList->front());
                channel.broadcast(idProcess, std::make_shared<MessageDECIDE>(m->common_est));
                if (_debugLevel >= 3) {
                    std::stringstream s;
                    s << "P" << idProcess << ": broadcast DECIDE(common_est=" << m->common_est << ")\n";
                    std::cerr << s.str();
                    s.str(std::string());
                }
                int v = m->common_est;
                Channel::freeMessageList(messageList);
                return v;
            }
        }
        // Line 22
        if (_debugLevel >= 20) {
            std::stringstream s;
            s << "P" << idProcess << ": coin.toss()\n";
            std::cerr << s.str();
        }
        c = coin.toss(idProcess, failureType);
        // Lines 23-25
        {
            bool unanimity = true;
            for (auto const& m : *messageList) {
                std::shared_ptr<const MessageAGREE> m2 = std::static_pointer_cast<const MessageAGREE>(m);
                if (m2->same) {
                    unanimity = false;
                    break;
                }
            }
            if (unanimity)
                est = c;
        }
        Channel::freeMessageList(messageList);
    } while (1);
}


MessageList* RandomConsensus::receive(unsigned int howMany, MessageFilter& filter)
{
    MessageList* messageList;

    messageList = new MessageList;
    do {
        std::shared_ptr<Message> message = channel.receive(idProcess, filter, true);
        if (message != 0) {
            if (_debugLevel >= 12) {
                std::stringstream s;
                s << "P" << idProcess << ": received message of type " << message->type << '\n';
                std::cerr << s.str();
                s.str(std::string());
            }
            if (message->type == MessageDECIDE::type)
                decide = std::static_pointer_cast<MessageDECIDE>(message);
            else
                messageList->push_back(message);
        }
    } while (messageList->size() < howMany && decide == 0);
    return messageList;
}


MessagePROP::MessagePROP(unsigned int _r, int _est) :
    Message(MessagePROP::type),
    r(_r),
    est(_est)
{
}


MessagePROP::~MessagePROP()
{
}


MessageCHECK::MessageCHECK(unsigned int _r, int _common_est) :
    Message(MessageCHECK::type),
    r(_r),
    common_est(_common_est)
{
}


MessageCHECK::~MessageCHECK()
{
}


MessageAGREE::MessageAGREE(unsigned int _r, int _common_est, bool _same) :
    Message(MessageAGREE::type),
    r(_r),
    common_est(_common_est),
    same(_same)
{
}


MessageAGREE::~MessageAGREE()
{
}


MessageDECIDE::MessageDECIDE(int _v) :
    Message(MessageDECIDE::type),
    v(_v)
{
}


MessageDECIDE::~MessageDECIDE()
{
}
