/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#ifndef _RANDOMCONSENSUS_H_
#define _RANDOMCONSENSUS_H_

#include "Consensus.h"
#include "Channel.h"

class MessagePROP : public Message {
public:
    MessagePROP(unsigned int r, int est);
    ~MessagePROP();
    static const int type = 0x00000001;
    const unsigned int r;
    const int est;
};


class MessageCHECK : public Message {
public:
    MessageCHECK(unsigned int r, int common_est);
    ~MessageCHECK();
    static const int type = 0x00000002;
    const unsigned int r;
    const int common_est;
};


class MessageAGREE : public Message {
public:
    MessageAGREE(unsigned int r, int common_est, bool same);
    ~MessageAGREE();
    static const int type = 0x00000003;
    const unsigned int r;
    const int common_est;
    const bool same;
};


class MessageDECIDE : public Message {
public:
    MessageDECIDE(int v);
    ~MessageDECIDE();
    static const int type = 0x00000004;
    const int v;
};


struct RandomConsensusStatistics {
    static const int MAX_R = 100;
    static const int MAX_V = 2;
    RandomConsensusStatistics(unsigned int n);
    RandomConsensusStatistics(unsigned int n, unsigned int f);
    unsigned int numExecutions, numCorrectExecutions;
    unsigned int n, f;
    ObservedVariable<unsigned int> numRounds;
    unsigned proposed[MAX_V]; /**< Counter per proposed values. */
    unsigned decided[MAX_V]; /**< Counter per decided values. */
    RandomConsensusStatistics& operator += (const RandomConsensusStatistics& other);
    friend std::ostream& operator << (std::ostream& out, const RandomConsensusStatistics& statistics);
    void outputXml(bool xml) { this->xml = xml; };
private:
    bool xml;
};


class RandomConsensus : public Consensus {
public:
	RandomConsensus(unsigned int numProcesses, int idProcess, Channel& channel, CommonCoin& coin, unsigned int failureType);
	virtual ~RandomConsensus();
	virtual int propose(int v);
    RandomConsensusStatistics statistics;
protected:
	class FilterPROP;
	class FilterCHECK;
	class FilterAGREE;
    Channel& channel;
	CommonCoin& coin;
	const unsigned int failureType;
	unsigned int r;
	int est, common_est, c; // Only 0-1 actually used, as this is binary consensus
	bool same;
	std::shared_ptr<const MessageDECIDE> decide;
	int task1();
    MessageList* receive(unsigned int howMany, MessageFilter& filter);
};


class RandomConsensus::FilterPROP : public MessageFilter {
public:
    FilterPROP(unsigned int r, std::shared_ptr<const MessageDECIDE>* decide);
    bool operator()(std::shared_ptr<const Message> mensaje);
private:
    const unsigned int r;
    std::shared_ptr<const MessageDECIDE>* const decide;
};


class RandomConsensus::FilterCHECK : public MessageFilter {
public:
    FilterCHECK(unsigned int r, std::shared_ptr<const MessageDECIDE>* _decide);
    bool operator()(std::shared_ptr<const Message> mensaje);
private:
    const unsigned int r;
    std::shared_ptr<const MessageDECIDE>* const decide;
};


class RandomConsensus::FilterAGREE : public MessageFilter {
public:
    FilterAGREE(unsigned int r, std::shared_ptr<const MessageDECIDE>* _decide);
    bool operator()(std::shared_ptr<const Message> mensaje);
private:
    const unsigned int r;
    std::shared_ptr<const MessageDECIDE>* const decide;;
};


#endif
