/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#include <sys/types.h>      // getrusage
#include <sys/time.h>       // getrusage
#include <sys/resource.h>   // getrusage
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <string>
#include <iostream>
#include <fstream>              // ifstream
#include <sstream>
#include <thread>
#include <algorithm>            // max
#include "Arguments.h"
#include "Coin.h"

using namespace std;


struct WorkData {
    ofstream outputDataFile;
    ofstream outputCommandsFile;
    std::mutex workerMutex;
};


void coinTestThread(int idProcess, unsigned int numSimulations, CommonCoin* coin, unsigned int failureType)
{
    stringstream s;
    int v;

    for (unsigned int i = 0; i < numSimulations; i++) {
        v = coin->toss(idProcess, failureType);
        if (coin->getDebug() > 0)
            cout << "P" << idProcess << ",T" << i+1 << ": TOSS=" << v << '\n';
        if (programArguments.showProgress && idProcess == (int) programArguments.nMin-1 && (numSimulations < 500 || (i % 50) == 0))
            cerr << '.' << flush;
    }
    if (programArguments.showProgress && idProcess == (int) programArguments.nMin-1)
        cerr << "\n\n" << flush;

}


void test_array_coin(unsigned int n, unsigned int f, unsigned int numSimulations, CommonCoinStatistics* testStatistics)
{
    int results[n];
    ArrayCoin coin(n, f);

    if (programArguments.showProgress) {
        std::stringstream s;
        s << "n=" << n << ",f=" << f << '\n';
        cerr << s.str() << flush;
    }
    coin.debug(programArguments.debugLevel);

    for (unsigned int i = 0; i < numSimulations; i++) {
        coin.toss(results);
        testStatistics->countToss(results);
        if (programArguments.showProgress && (numSimulations < 500 || (i % 50) == 0))
            cerr << '.' << flush;
    }
    if (programArguments.showProgress)
        cerr << "\n\n" << flush;
}


void test_synthetic_coin(unsigned int n, unsigned int f, unsigned int numSimulations, CommonCoinStatistics* testStatistics)
{
    double bias;
    unsigned int fm;

    bias = 0;
    do {
        if (programArguments.showProgress) {
            std::stringstream s;
            s << "n=" << n << ",f=" << f << ",bias=" << bias << '\n';
            cerr << s.str() << flush;
        }

        SyntheticCommonCoin coin(n, f, bias);
        thread* threads[n];

        coin.debug(programArguments.debugLevel);

        // Launches threads
        for (unsigned int i = 0; i < n; i++) {
            fm = (programArguments.failureType != ProgramArguments::FAILURE_NO && i < f) ? programArguments.failureType : ProgramArguments::FAILURE_NO;
            threads[i] = new thread(coinTestThread, i, numSimulations, &coin, fm);
        }

        for (unsigned int i = 0; i < n; i++) {
            threads[i]->join();
            delete threads[i];
        }

        bias += 0.25;
        cout << "\n\n";
    } while (bias <= 1);
}


void test_simulated_coin(unsigned int n, unsigned int f, unsigned int numSimulations, CommonCoinStatistics* testStatistics)
{
	Channel network;
    SimulatedCommonCoin coin(n, f, network);
    thread* threads[n];
    unsigned int fm;

	if (programArguments.showProgress) {
        std::stringstream s;
        s << "n=" << n << ",f=" << f << '\n';
        cerr << s.str() << flush;
    }

    coin.debug(programArguments.debugLevel);
    network.debug(programArguments.debugLevel);

    for (unsigned int i = 0; i < n; i++) {
		network.attach(i);
	}

	// Choose randomly which f processes will fail
    bool failedProcesses[n];
    for (unsigned int i = 0; i < n; i++)
        failedProcesses[i] = false;
    for (unsigned int i = 0; i < f; i++) {
        unsigned int paux;
        do {
            paux = random() % n;
        } while (failedProcesses[paux]);
        failedProcesses[paux] = true;
    }

    // Launches threads
    for (unsigned int i = 0; i < n; i++) {
        fm = (programArguments.failureType != ProgramArguments::FAILURE_NO && failedProcesses[i]) ? programArguments.failureType : ProgramArguments::FAILURE_NO;
        threads[i] = new thread(coinTestThread, i, numSimulations, &coin, fm);
    }

    for (unsigned int i = 0; i < n; i++) {
        threads[i]->join();
        delete threads[i];
    }

    cout << coin.statistics << '\n';
    cout << "\n\n";
    cout << network.statistics << '\n';

    cout << "\n\n";

    *testStatistics = coin.statistics;
}


// Rough estimation of the compute demand for a simulation with n processes, where f are faulty processes
static double estimateComputeDemand(int n, int f)
{
	return 3.0*pow(1.0*n-f/5.0,2.5);
}


static double estimateTotalComputeDemand()
{
    double demand;
    unsigned int n, f, ndelta, f_ant, pf;

    demand = 0;
    ndelta = max<unsigned int>(1,(programArguments.nMax - programArguments.nMin) / programArguments.numXPoints);
    n = programArguments.nMin;
    do {
        pf = programArguments.pfMin;
        f_ant = 0xffffffff;
        do {
            f = (int) ((double) n * pf / 100.0);
            f = std::min<unsigned int>(f, n-(n/2+1));
            if (f != f_ant) {
                demand += estimateComputeDemand(n,f);
                f_ant = f;
            }
            pf += programArguments.pfStep;
        } while (pf <= programArguments.pfMax);
        if (n < programArguments.nMax) {
            n += ndelta;
            if (n > programArguments.nMax)
                n = programArguments.nMax;
        }
        else
            n += ndelta;
    } while (n <= programArguments.nMax);
    return demand;
}


void performExperiment(WorkData* workData)
{
    unsigned int n, f, ndelta, f_ant, pf;
    // Computes the best number of simulation threads to use:
    // To begin with, it will be equal to the hardware concurrency value given by the C++ runtime. If that is zero (unknown), it will use 1 thread only.
    // Second, it will be not higher than the number of simulations to be run. This is to take into consideration simulations with low repetition counts.
    unsigned int numSimulationThreads = std::min<unsigned int>(std::max<unsigned int>(std::thread::hardware_concurrency(), 1), programArguments.numSimulations);
    double cumulativeCpuTime, estimatedTotalComputeDemand, estimatedConsumedComputeDemand;
    struct rusage ru1, ru2;

    std::cerr << "INFO: using " << numSimulationThreads << " parallel simulation threads.\n\n";

    ndelta = max<unsigned int>(1,(programArguments.nMax - programArguments.nMin) / programArguments.numXPoints);

    estimatedTotalComputeDemand = estimateTotalComputeDemand();
    estimatedConsumedComputeDemand = 0;
    cumulativeCpuTime = 0;
    n = programArguments.nMin;
    getrusage(RUSAGE_SELF, &ru1);
    do {

        pf = programArguments.pfMin;
        f_ant = 0xffffffff;
        do {
            f = (int) ((double) n * pf / 100.0);
            f = std::min<unsigned int>(f, n-(n/2+1));
            if (f != f_ant) {
                CommonCoinStatistics testStatistics(n,f);
                unsigned int numSimulationsPerThread = programArguments.numSimulations / numSimulationThreads;
                thread* simulationThreads[numSimulationThreads];
                CommonCoinStatistics* testStatisticsAux[numSimulationThreads];

                for (unsigned int ti = 0; ti < numSimulationThreads; ti++) {
                    testStatisticsAux[ti] = new CommonCoinStatistics(n,f);
//	                simulationThreads[ti] = new thread(test_array_coin, n, f, numSimulationsPerThread, testStatisticsAux[ti]);
//	                simulationThreads[ti] = new thread(test_synthetic_coin, n, f, numSimulationsPerThread, testStatisticsAux[ti]);
                    simulationThreads[ti] = new thread(test_simulated_coin, n, f, numSimulationsPerThread, testStatisticsAux[ti]);
                }
                for (unsigned int ti = 0; ti < numSimulationThreads; ti++) {
                    simulationThreads[ti]->join();
                    testStatistics += *testStatisticsAux[ti];
                    delete testStatisticsAux[ti];
                    delete simulationThreads[ti];
                }
                workData->outputDataFile << n << '\t' << f << '\t' << pf << '\t' << (double) testStatistics.numCommonTosses / testStatistics.numTosses * 100.0 << '\t' << (double) testStatistics.commonValue[0] / testStatistics.numCommonTosses * 100.0 << '\t' << (double) testStatistics.commonValue[1] / testStatistics.numCommonTosses * 100.0 << '\n';
                workData->outputDataFile << flush;

                estimatedConsumedComputeDemand += estimateComputeDemand(n,f);
        		getrusage(RUSAGE_SELF, &ru2);
        		cumulativeCpuTime += (ru2.ru_utime.tv_sec + ru2.ru_stime.tv_sec - ru1.ru_utime.tv_sec - ru1.ru_stime.tv_sec + (ru2.ru_utime.tv_usec + ru2.ru_stime.tv_usec - ru1.ru_utime.tv_usec - ru1.ru_stime.tv_usec) / 1e6);
        		ru1 = ru2;
        		if (estimatedConsumedComputeDemand > 0 && programArguments.showProgress && estimatedConsumedComputeDemand / estimatedTotalComputeDemand > 0.01) {
                    double estimatedRemainingCpuTime = (1 - estimatedConsumedComputeDemand / estimatedTotalComputeDemand) / (estimatedConsumedComputeDemand / estimatedTotalComputeDemand) * cumulativeCpuTime;
                    std::cerr << "CPU INFO >>> used: " << cumulativeCpuTime << "s (" << estimatedConsumedComputeDemand / estimatedTotalComputeDemand * 100.0 << "% total estimated), estimated remaining: " << estimatedRemainingCpuTime << "s. Estimated total: " << cumulativeCpuTime + estimatedRemainingCpuTime << "s.\n\n";
        		}
                f_ant = f;
            }
            pf += programArguments.pfStep;
        } while (pf <= programArguments.pfMax);
        if (n < programArguments.nMax) {
            n += ndelta;
            if (n > programArguments.nMax)
                n = programArguments.nMax;
        }
        else
            n += ndelta;
    } while (n <= programArguments.nMax);
}


int main(int argc, char** argv)
{
    WorkData workData;

	readArguments(argc, argv, programArguments);
    workData.outputDataFile.open(programArguments.outputFilename + ".data");
    if (! workData.outputDataFile.is_open()) {
        cerr << "ERROR: cannot create " << programArguments.outputFilename + ".data" << '\n';
        return EX_CANTCREAT;
    }
    workData.outputCommandsFile.open(programArguments.outputFilename + ".r");
    if (! workData.outputCommandsFile.is_open()) {
        cerr << "ERROR: cannot create " << programArguments.outputFilename + ".r" << '\n';
        return EX_CANTCREAT;
    }

	srandom(time(NULL) * getpid());

    workData.outputDataFile << "# Experiments for: Common Coin Toss\n#\n";
    workData.outputDataFile << "# n_min = " << programArguments.nMin << '\n';
    workData.outputDataFile << "# n_max = " << programArguments.nMax << '\n';
    workData.outputDataFile << "# num_simulations = " << programArguments.numSimulations << '\n';
    workData.outputDataFile << "# failure_type = " << failureTypeName[programArguments.failureType] << '\n';
    workData.outputDataFile << "# coin_prob0 = \"" << coinProb0 << "\"\n";
    workData.outputDataFile << programArguments.comments;

    workData.outputDataFile << "n\tf\tpf\tpCommon\tp0\tp1\n";

    performExperiment(&workData);

//    cout << "plot(x,z,type=\"o\",xlab=\"Number of processes\",ylab=\"Number of messages\",main=\"Processes - Messages\")\n";
    workData.outputCommandsFile << programArguments.comments;
    workData.outputCommandsFile << "n_min <- " << programArguments.nMin << '\n';
    workData.outputCommandsFile << "n_max <- " << programArguments.nMax << '\n';
    workData.outputCommandsFile << "num_simulations <- " << programArguments.numSimulations << '\n';
    workData.outputCommandsFile << "failure_type <- \"" << failureTypeName[programArguments.failureType] << "\"\n";
    workData.outputCommandsFile << "coin_prob0 <- \"" << coinProb0 << "\"\n";
    workData.outputCommandsFile << "\n\n";
    workData.outputCommandsFile << "df <- read.table(\"" << programArguments.outputFilename + ".data" << "\",header=TRUE)\n";
    workData.outputCommandsFile << "\n\n";
    workData.outputCommandsFile << "summary(df)\n";
    workData.outputCommandsFile << "\n\n";

    unsigned int col = 1;
    stringstream dataLegend;
    stringstream plotSubtitle;
    for (unsigned pf = programArguments.pfMin; pf <= programArguments.pfMax; pf += programArguments.pfStep, col++) {
        if (pf == programArguments.pfMin) {
            workData.outputCommandsFile << "dev.new()\n";
            plotSubtitle << "# Simulations: " << programArguments.numSimulations << ", Failures: " << failureTypeName[programArguments.failureType];
            plotSubtitle << ", P(0): " << coinProb0;
            if (programArguments.plotSubtitle.size() > 0)
                plotSubtitle << ", " << programArguments.plotSubtitle;
            workData.outputCommandsFile << "plot(0,type=\"n\",xlim=c(min(df$n),max(df$n)),ylim=c(0,100),main=\"pCommon - Processes\",xlab=\"# Processes\",ylab=\"p Common\",sub=\"" << plotSubtitle.str() << "\")\n";
        }
        else
            dataLegend << ',';
        workData.outputCommandsFile << "x <- unique(subset(df,pf==" << pf << "))$n\n";
        workData.outputCommandsFile << "y <- unique(subset(df,pf==" << pf << "))$pCommon\n";
        workData.outputCommandsFile << "p <- matrix(c(x,y),ncol=2)\n";
        workData.outputCommandsFile << "lines(p,col=palette()[" << col << "],lty=1)\n";
        workData.outputCommandsFile << "points(p,col=palette()[" << col << "],pch=" << col << ")\n";
        workData.outputCommandsFile << "y <- unique(subset(df,pf==" << pf << "))$p0\n";
        workData.outputCommandsFile << "p <- matrix(c(x,y),ncol=2)\n";
        workData.outputCommandsFile << "lines(p,col=palette()[" << col << "],lty=2)\n";
        workData.outputCommandsFile << "points(p,col=palette()[" << col << "],pch=" << col << ")\n";
        dataLegend << "\"pf=" << pf << "\"";

    }
    workData.outputCommandsFile << "legend(x=\"bottomleft\",legend=c(" << dataLegend.str() << "),col=palette(),lwd=1)\n";
    workData.outputCommandsFile << "legend(x=\"bottomleft\",legend=rep(\"\",each=" << col-1 << "),col=palette(),lwd=1,pch=c(1:" << col-1 << "),bty='n')\n";

/*
    workData.outputCommandsFile << "dev.new()\n";
    workData.outputCommandsFile << "plot(pCommon ~ n,data=unique(subset(df,pf==0)),type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(0,100),main=\"pCommon - Processes\",xlab=\"# Processes\",ylab=\"p Common\",col=palette()[1])\n";
    workData.outputCommandsFile << "par(new=TRUE); par(ann=FALSE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
    workData.outputCommandsFile << "plot(p0 ~ n,data=unique(subset(df,pf==0)),type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(0,100),col=palette()[1],lty=2)\n";
    workData.outputCommandsFile << "plotLegend <- paste(\"Colors for pf:   0=\",palette()[1],sep=\"\")\n";
    for (unsigned pf = programArguments.pfMin; pf <= programArguments.pfMax; pf += programArguments.pfStep) {
        workData.outputCommandsFile << "plotLegend <- paste(plotLegend,\", " << pf << "=\", palette()[" << pf/10+1 << "],sep=\"\")\n";
        workData.outputCommandsFile << "par(new=TRUE); par(ann=FALSE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
        workData.outputCommandsFile << "plot(pCommon ~ n,data=unique(subset(df,pf==" << pf << ")),type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(0,100),col=palette()[" << pf/10+1 << "])\n";
        workData.outputCommandsFile << "par(new=TRUE); par(ann=FALSE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
        workData.outputCommandsFile << "plot(p0 ~ n,data=unique(subset(df,pf==" << pf << ")),type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(0,100),col=palette()[" << pf/10+1 << "],lty=2)\n";
    }
    workData.outputCommandsFile << "par(new=TRUE); par(ann=TRUE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
    workData.outputCommandsFile << "plot(0,0,type=\"n\",sub=plotLegend,xlab=\"\",ylab=\"\")\n";
*/

    workData.outputCommandsFile << "dev.print(tiff,filename=\"" << programArguments.outputFilename << ".tif\",width=650,compression=\"lzw\",bg=\"white\")\n";

    workData.outputDataFile.close();
    workData.outputCommandsFile.close();


	return EX_OK;

//	test_vectorclock();
//	exit(0);

//	test_channels(n, f, numSimulations);
//	exit(0);


}
