/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <string>
#include <iostream>
#include <fstream>              // ifstream
#include <sstream>
#include <thread>
#include <algorithm>            // max
#include "Arguments.h"
#include "DRConsensus.h"
#include "Channel.h"

using namespace std;


struct WorkData {
    ofstream outputDataFile;
    ofstream outputCommandsFile;
    ofstream outputExtendedDataFile;
    std::mutex workerMutex;
};


void processThread(DRConsensus* consensus)
{
    int v;
    int idProcess;
    stringstream s;

    idProcess = consensus->idProcess;
    v = random() % 2;
    if (consensus->getDebug() >= 3) {
        s << "P" << idProcess << ": PROPOSAL=" << v << '\n';
        cerr << s.str();
        s.str(std::string());
    }
    v = consensus->propose(v);
    if (consensus->getDebug() >= 3) {
        s << "P" << idProcess << ": CONSENSUATED=" << v << '\n';
        cerr << s.str();
        s.str(std::string());
    }
}


void simulationThread(unsigned int n, unsigned int f, unsigned int pf, double coinBias, unsigned int numUnstableRounds, unsigned int numSimulations, DRConsensusStatistics* consensusStatistics, ChannelStatistics* networkStatistics, WorkData* workData)
{
	Channel network;
	DRConsensus* consensus[n];
	thread* threads[n];
	unsigned int auxNumMessagesSent;

	if (programArguments.showProgress) {
        std::stringstream s;
        s << "n=" << n << ",f=" << f << ",bias=" << coinBias << ",numUnstableRounds=" << numUnstableRounds << '\n';
        cerr << s.str() << flush;
    }

	for (unsigned int i = 0; i < n; i++) {
		network.attach(i);
	}

	for (unsigned int s = 0; s < numSimulations; s++) {
        if (programArguments.showProgress && (numSimulations < 500 || (s % 50) == 0))
            cerr << '.' << flush;
        network.reset();
        auxNumMessagesSent = network.statistics.numMessagesSent.sum();
        SyntheticFailureDetector failureDetector(n, f, numUnstableRounds);
        SyntheticCommonCoin coin(n, f, coinBias);
        // SimulatedCommonCoin coin(n, f, network);
        coin.debug(programArguments.debugLevel);
        for (unsigned int i = 0; i < n; i++) {
            if (i < f)
                failureDetector.mayFail(i);
            consensus[i] = new DRConsensus(n, i, network, coin, failureDetector, i < f ? ProgramArguments::FAILURE_START : ProgramArguments::FAILURE_NO);
            consensus[i]->debug(programArguments.debugLevel);
            threads[i] = new thread(processThread, consensus[i]);
        }
        for (unsigned int i = 0; i < n; i++)
            threads[i]->join();
        network.collectStatistics();
        for (unsigned int i = 0; i < n; i++) {
            {
                std::unique_lock<std::mutex> l0(workData->workerMutex);
                workData->outputExtendedDataFile << n << '\t' << f << '\t' << pf << '\t' << coinBias << '\t' << numUnstableRounds << '\t' << (network.statistics.numMessagesSent.sum() - auxNumMessagesSent) / n << '\t' << consensus[i]->statistics.numRounds.sum() << '\n';
                workData->outputExtendedDataFile << flush;
            }
            *consensusStatistics += consensus[i]->statistics;
            delete consensus[i];
            delete threads[i];
        }
	}
	if (programArguments.showProgress)
        cerr << '\n' << flush;


    *networkStatistics = network.statistics;
    if (programArguments.debugLevel > 0) {
        stringstream s;
        consensusStatistics->outputXml(false);
        s << *consensusStatistics << '\n';
        s << *networkStatistics << '\n';
        cerr << s.str() << flush;
    }
}

// double coinBias[] = { 0, 0.25, 0.5, 0.75, 1.0, -1};
double coinBias[] = { 0, 0.75, -1};
int numUnstableRounds[] = { 0, 1, 2, 5, 1000, -1};

void performExperiment(WorkData* workData)
{
	unsigned int n, f, ndelta, f_ant, cb, pf, nur;
    // Computes the best number of simulation threads to use:
    // To begin with, it will be equal to the hardware concurrency value given by the C++ runtime. If that is zero (unknown), it will use 1 thread only.
    // Second, it will be not higher than the number of simulations to be run. This is to take into consideration simulations with low repetition counts.
    unsigned int numSimulationThreads = std::min<unsigned int>(std::max<unsigned int>(std::thread::hardware_concurrency(), 1), programArguments.numSimulations);

    ndelta = max<unsigned int>(1,(programArguments.nMax - programArguments.nMin) / programArguments.numXPoints);

    n = programArguments.nMin;
    do {
        cb = 0;
        do {
            nur = 0;
            do {
				// Skips the case where both coin and failure detector are useless
				if (! (coinBias[cb] == 0 && numUnstableRounds[nur] > 5)) {
                pf = programArguments.pfMin;
                f_ant = 0xffffffff;
                do {
                    f = (int) ((double) n * pf / 100.0);
                    f = std::min<unsigned int>(f, n-(n/2+1));
                    if (f != f_ant) {
                        DRConsensusStatistics consensusStatistics(n, f);
                        ChannelStatistics networkStatistics;
                        thread* simulationThreads[numSimulationThreads];
                        DRConsensusStatistics* consensusStatisticsAux[numSimulationThreads];
                        ChannelStatistics* networkStatisticsAux[numSimulationThreads];
                        unsigned int numSimulationsPerThread = programArguments.numSimulations / numSimulationThreads;

                        for (unsigned int ti = 0; ti < numSimulationThreads; ti++) {
                            consensusStatisticsAux[ti] = new DRConsensusStatistics(n, f);
                            networkStatisticsAux[ti] = new ChannelStatistics;
                            simulationThreads[ti] = new thread(simulationThread, n, f, pf, coinBias[cb], numUnstableRounds[nur], numSimulationsPerThread, consensusStatisticsAux[ti], networkStatisticsAux[ti], workData);
                        }
                        for (unsigned int ti = 0; ti < numSimulationThreads; ti++) {
                            simulationThreads[ti]->join();
                            consensusStatistics += *consensusStatisticsAux[ti];
                            networkStatistics += *networkStatisticsAux[ti];
                            delete consensusStatisticsAux[ti];
                            delete networkStatisticsAux[ti];
                            delete simulationThreads[ti];
                        }
                        // TODO: the computation of the average number of messages is not clear: do we divide just by correct executions? Do we substract messages for failed processes in the statistics (would require counting separately per process)? Etc.
                        workData->outputDataFile << n << '\t' << f << '\t' << pf << '\t' << coinBias[cb] << '\t' << numUnstableRounds[nur] << '\t' << networkStatistics.numMessagesSent.average() << '\t' << networkStatistics.numMessagesSent.stdDev() << '\t' << consensusStatistics.numRounds.average() << '\t' << consensusStatistics.numRounds.stdDev() << '\n';
                        workData->outputDataFile << flush;
                        f_ant = f;
                    }
                    pf += programArguments.pfStep;
                } while (pf <= programArguments.pfMax);
				}
                nur++;
            } while (numUnstableRounds[nur] >= 0);
            cb++;
        } while (coinBias[cb] >= 0);
        if (n < programArguments.nMax) {
            n += ndelta;
            if (n > programArguments.nMax)
                n = programArguments.nMax;
        }
        else
            n += ndelta;
    } while (n <= programArguments.nMax);
}




int main(int argc, char** argv)
{
    WorkData workData;
    unsigned int nur;
//    unsigned int cb, nur;

	readArguments(argc, argv, programArguments);
    workData.outputDataFile.open(programArguments.outputFilename + ".datakkJMR");
    if (! workData.outputDataFile.is_open()) {
        cerr << "ERROR: cannot create " << programArguments.outputFilename + ".data" << '\n';
        return EX_CANTCREAT;
    }
    workData.outputCommandsFile.open(programArguments.outputFilename + ".r");
    if (! workData.outputCommandsFile.is_open()) {
        cerr << "ERROR: cannot create " << programArguments.outputFilename + ".r" << '\n';
        return EX_CANTCREAT;
    }
    workData.outputExtendedDataFile.open(programArguments.outputFilename + "_extended.data");
    if (! workData.outputExtendedDataFile.is_open()) {
        cerr << "ERROR: cannot create " << programArguments.outputFilename + "_extended.data" << '\n';
        return EX_CANTCREAT;
    }

	srandom(time(NULL) * getpid());

	{
        stringstream s;

        s << "# Experiments for: Random Consensus with Coin\n#\n";
        s << "# n_min = " << programArguments.nMin << '\n';
        s << "# n_max = " << programArguments.nMax << '\n';
        s << "# num_simulations = " << programArguments.numSimulations << '\n';
        s << "# failure_type = " << failureTypeName[programArguments.failureType] << '\n';
        s << "# coin_prob0 = \"" << coinProb0 << "\"\n";
        s << programArguments.comments;
        workData.outputDataFile << s.str();
        workData.outputExtendedDataFile << s.str();
	}

    workData.outputDataFile << "n\tf\tpf\tbias\tnumUnstableRounds\tnumMessages\tdevMessages\tavgRounds\tdevRounds\n";
    workData.outputExtendedDataFile << "n\tf\tpf\tbias\tnumUnstableRounds\tnumMessages\tavgRounds\n";

//    performExperiment(&workData);

//    cout << "plot(x,z,type=\"o\",xlab=\"Number of processes\",ylab=\"Number of messages\",main=\"Processes - Messages\")\n";
    workData.outputCommandsFile << programArguments.comments;
    workData.outputCommandsFile << "n_min <- " << programArguments.nMin << '\n';
    workData.outputCommandsFile << "n_max <- " << programArguments.nMax << '\n';
    workData.outputCommandsFile << "num_simulations <- " << programArguments.numSimulations << '\n';
    workData.outputCommandsFile << "failure_type <- \"" << failureTypeName[programArguments.failureType] << "\"\n";
    workData.outputCommandsFile << "coin_prob0 <- \"" << coinProb0 << "\"\n";
    workData.outputCommandsFile << "\n\n";
    workData.outputCommandsFile << "df <- read.table(\"" << programArguments.outputFilename + ".data" << "\",header=TRUE)\n";
    workData.outputCommandsFile << "\n\n";
    workData.outputCommandsFile << "summary(df)\n";
    workData.outputCommandsFile << "\n\n";

{
    unsigned int col = 1;
    stringstream dataLegend;
    stringstream plotSubtitle;
    for (unsigned pf = programArguments.pfMin; pf <= programArguments.pfMax; pf += programArguments.pfStep, col++) {
        if (pf == programArguments.pfMin) {
            workData.outputCommandsFile << "dev.new()\n";
            workData.outputCommandsFile << "par(mar=c(5,5,1,1))\n";
            plotSubtitle << "# Simulations: " << programArguments.numSimulations << ", Failures: " << failureTypeName[programArguments.failureType];
            plotSubtitle << ", P(0): " << coinProb0;
            if (programArguments.plotSubtitle.size() > 0)
                plotSubtitle << ", " << programArguments.plotSubtitle;
            workData.outputCommandsFile << "plot(0,type=\"n\",xlim=c(min(df$n),max(df$n)),ylim=c(min(df$avgRounds-df$devRounds/10),max(df$avgRounds+df$devRounds/10)),cex.axis=1.5,cex.lab=1.5,xlab=\"# Processes\",ylab=\"Avg # Rounds\")\n";
        }
        else
            dataLegend << ',';
        nur = 0;
        do {
            workData.outputCommandsFile << "x <- unique(subset(df,pf==" << pf << " & bias==0 & numUnstableRounds==" << numUnstableRounds[nur] << "))$n\n";
            workData.outputCommandsFile << "y <- unique(subset(df,pf==" << pf << " & bias==0 & numUnstableRounds==" << numUnstableRounds[nur] << "))$avgRounds\n";
            workData.outputCommandsFile << "s <- unique(subset(df,pf==" << pf << " & bias==0 & numUnstableRounds==" << numUnstableRounds[nur] << "))$devRounds / 20\n";
            workData.outputCommandsFile << "p <- matrix(c(x,y),ncol=2)\n";
            workData.outputCommandsFile << "sampleWidth <- (max(x) - min(x)) / 100\n";
            workData.outputCommandsFile << "lines(p,col=palette()[" << col << "])\n";
            workData.outputCommandsFile << "points(p,col=palette()[" << col << "],pch=" << nur+1 << ")\n";
            if (nur == 0)
                dataLegend << "\"pf=" << pf << "\"";
//            workData.outputCommandsFile << "segments(x-sampleWidth,y-s,x-sampleWidth,y+s,col=palette()[" << col << "])\n";
//            workData.outputCommandsFile << "segments(x+sampleWidth,y-s,x+sampleWidth,y+s,col=palette()[" << col << "])\n";
//            workData.outputCommandsFile << "segments(x-sampleWidth,y-s,x+sampleWidth,y-s,col=palette()[" << col << "])\n";
//            workData.outputCommandsFile << "segments(x-sampleWidth,y+s,x+sampleWidth,y+s,col=palette()[" << col << "])\n";
            nur++;
        } while (numUnstableRounds[nur] < 1000);
    }
    workData.outputCommandsFile << "legend(x=\"topright\",inset=c(0.1,0.2),legend=c(" << dataLegend.str() << "),col=palette(),lwd=1)\n";
    nur = 0;
    stringstream fdLegend;
    do {
        if (nur > 0)
            fdLegend << ',';
        fdLegend << '"' << "FD " << numUnstableRounds[nur] << '"';
        nur++;
    } while (numUnstableRounds[nur] < 1000);
    workData.outputCommandsFile << "legend(x=\"topright\",inset=c(0.4,0.2),legend=c(" << fdLegend.str() << "),col=1,lwd=1,pch=c(1:" << nur << "))\n";

    workData.outputCommandsFile << "dev.print(png,filename=\"" << programArguments.outputFilename << "c.png\",width=1050,bg=\"white\")\n";
    // convert -background white -alpha remove orig.tif new.tif
}

    numUnstableRounds[0] = 1000;
    numUnstableRounds[1] = -1;
    numUnstableRounds[2] = 1000;
    numUnstableRounds[3] = -1;

{
    unsigned int col = 1;
    stringstream dataLegend;
    stringstream plotSubtitle;
    for (unsigned pf = programArguments.pfMin; pf <= programArguments.pfMax; pf += programArguments.pfStep, col++) {
        if (pf == programArguments.pfMin) {
            workData.outputCommandsFile << "dev.new()\n";
            workData.outputCommandsFile << "par(mar=c(5,5,1,1))\n";
            plotSubtitle << "# Simulations: " << programArguments.numSimulations << ", Failures: " << failureTypeName[programArguments.failureType];
            plotSubtitle << ", P(0): " << coinProb0;
            if (programArguments.plotSubtitle.size() > 0)
                plotSubtitle << ", " << programArguments.plotSubtitle;
            workData.outputCommandsFile << "plot(0,type=\"n\",xlim=c(min(df$n),max(df$n)),ylim=c(min(df$avgRounds-df$devRounds/10),2.5),cex.axis=1.5,cex.lab=1.5,xlab=\"# Processes\",ylab=\"Avg # Rounds\")\n";
        }
        else
            dataLegend << ',';
        nur = 0;
        do {
            workData.outputCommandsFile << "x <- unique(subset(df,pf==" << pf << " & bias==0.75 & numUnstableRounds==" << numUnstableRounds[nur] << "))$n\n";
            workData.outputCommandsFile << "y <- unique(subset(df,pf==" << pf << " & bias==0.75 & numUnstableRounds==" << numUnstableRounds[nur] << "))$avgRounds\n";
            workData.outputCommandsFile << "s <- unique(subset(df,pf==" << pf << " & bias==0.75 & numUnstableRounds==" << numUnstableRounds[nur] << "))$devRounds / 20\n";
            workData.outputCommandsFile << "p <- matrix(c(x,y),ncol=2)\n";
            workData.outputCommandsFile << "sampleWidth <- (max(x) - min(x)) / 100\n";
            workData.outputCommandsFile << "lines(p,col=palette()[" << col << "])\n";
            workData.outputCommandsFile << "points(p,col=palette()[" << col << "],pch=" << nur+1 << ")\n";
            if (nur == 0)
                dataLegend << "\"pf=" << pf << "\"";
//            workData.outputCommandsFile << "segments(x-sampleWidth,y-s,x-sampleWidth,y+s,col=palette()[" << col << "])\n";
//            workData.outputCommandsFile << "segments(x+sampleWidth,y-s,x+sampleWidth,y+s,col=palette()[" << col << "])\n";
//            workData.outputCommandsFile << "segments(x-sampleWidth,y-s,x+sampleWidth,y-s,col=palette()[" << col << "])\n";
//            workData.outputCommandsFile << "segments(x-sampleWidth,y+s,x+sampleWidth,y+s,col=palette()[" << col << "])\n";
            nur++;
        } while (numUnstableRounds[nur] >= 0);
    }
    workData.outputCommandsFile << "legend(x=\"right\",inset=c(0.1,0.2),legend=c(" << dataLegend.str() << "),col=palette(),lwd=1)\n";
    nur = 0;
    stringstream fdLegend;
    do {
        if (nur > 0)
            fdLegend << ',';
        fdLegend << '"' << "FD " << numUnstableRounds[nur] << '"';
        nur++;
    } while (numUnstableRounds[nur] >= 0);
//    workData.outputCommandsFile << "legend(x=\"topright\",inset=c(0.4,0.2),legend=c(" << fdLegend.str() << "),col=1,lwd=1,pch=c(1:" << nur << "))\n";

    workData.outputCommandsFile << "dev.print(png,filename=\"" << programArguments.outputFilename << "fd.png\",width=1050,bg=\"white\")\n";
    // convert -background white -alpha remove orig.tif new.tif
}



    workData.outputDataFile.close();
    workData.outputExtendedDataFile.close();
    workData.outputCommandsFile.close();

    return EX_OK;
}
