/**********************************************************
 * @brief Simulation code for Binary Consensus in Anonymous Asynchronous Systems using Coins and Failure Detectors.
 * @date July 2018
 * @author Javier Martín Rueda, javier.martin@upm.es
 * @author José Luis López-Presa, joseluis.llopezp@upm.es
 * @author Ernesto Jiménez, ernes@etsisi.upm.es
 **********************************************************/


#include <sys/types.h>      // getrusage
#include <sys/time.h>       // getrusage
#include <sys/resource.h>   // getrusage
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <string>
#include <iostream>
#include <fstream>              // ifstream
#include <sstream>
#include <thread>
#include <algorithm>            // max
#include "Arguments.h"
#include "RandomConsensus.h"
#include "Channel.h"

using namespace std;


struct WorkData {
    ofstream outputDataFile;
    ofstream outputCommandsFile;
    ofstream outputExtendedDataFile;
    std::mutex workerMutex;
};


void processThread(RandomConsensus* consensus)
{
    int v;
    int idProcess;
    stringstream s;

    idProcess = consensus->idProcess;
    v = random() % 2;
    if (consensus->getDebug() >= 3) {
        s << "P" << idProcess << ": PROPOSAL=" << v << '\n';
        cerr << s.str();
        s.str(std::string());
    }
    v = consensus->propose(v);
    if (consensus->getDebug() >= 3) {
        s << "P" << idProcess << ": CONSENSUATED=" << v << '\n';
        cerr << s.str();
        s.str(std::string());
    }
}


void simulationThread(unsigned int n, unsigned int f, unsigned int pf, double coinBias, unsigned int numSimulations, RandomConsensusStatistics* consensusStatistics, ChannelStatistics* networkStatistics, WorkData* workData)
{
	Channel network;
	RandomConsensus* consensus[n];
	thread* threads[n];
	unsigned int auxNumMessagesSent;

	if (programArguments.showProgress) {
        std::stringstream s;
        s << "n=" << n << ",f=" << f << ",bias=" << coinBias << '\n';
        cerr << s.str() << flush;
    }

	for (unsigned int i = 0; i < n; i++) {
		network.attach(i);
	}

	for (unsigned int s = 0; s < numSimulations; s++) {
        if (programArguments.showProgress && (numSimulations < 500 || (s % 50) == 0))
            cerr << '.' << flush;
        network.reset();
        auxNumMessagesSent = network.statistics.numMessagesSent.sum();
        // SyntheticCommonCoin coin(n, f, coinBias);
        SimulatedCommonCoin coin(n, f, network);
        coin.debug(programArguments.debugLevel);
        for (unsigned int i = 0; i < n; i++) {
            consensus[i] = new RandomConsensus(n, i, network, coin, i < f ? ProgramArguments::FAILURE_START : ProgramArguments::FAILURE_NO);
            consensus[i]->debug(programArguments.debugLevel);
            threads[i] = new thread(processThread, consensus[i]);
        }
        for (unsigned int i = 0; i < n; i++)
            threads[i]->join();
        network.collectStatistics();
        for (unsigned int i = 0; i < n; i++) {
            {
                std::unique_lock<std::mutex> l0(workData->workerMutex);
                workData->outputExtendedDataFile << n << '\t' << f << '\t' << pf << '\t' << coinBias << '\t' << (network.statistics.numMessagesSent.sum() - auxNumMessagesSent) / n << '\t' << consensus[i]->statistics.numRounds.sum() << '\n';
                workData->outputExtendedDataFile << flush;
            }
            *consensusStatistics += consensus[i]->statistics;
            delete consensus[i];
            delete threads[i];
        }
	}
	if (programArguments.showProgress)
        cerr << '\n' << flush;


    *networkStatistics = network.statistics;
    if (programArguments.debugLevel > 0) {
        stringstream s;
        consensusStatistics->outputXml(false);
        s << *consensusStatistics << '\n';
        s << *networkStatistics << '\n';
        cerr << s.str() << flush;
    }
}


// Rough estimation of the compute demand for a simulation with n processes, where f are faulty processes
static double estimateComputeDemand(int n, int f)
{
	return 3.0*pow(1.0*n-f/5.0,2.5);
}


static double estimateTotalComputeDemand()
{
    double demand;
    unsigned int n, f, ndelta, f_ant, pf;

    demand = 0;
    ndelta = max<unsigned int>(1,(programArguments.nMax - programArguments.nMin) / programArguments.numXPoints);
    n = programArguments.nMin;
    do {
        pf = programArguments.pfMin;
        f_ant = 0xffffffff;
        do {
            f = (int) ((double) n * pf / 100.0);
            f = std::min<unsigned int>(f, n-(n/2+1));
            if (f != f_ant) {
                demand += estimateComputeDemand(n,f);
                f_ant = f;
            }
            pf += programArguments.pfStep;
        } while (pf <= programArguments.pfMax);
        if (n < programArguments.nMax) {
            n += ndelta;
            if (n > programArguments.nMax)
                n = programArguments.nMax;
        }
        else
            n += ndelta;
    } while (n <= programArguments.nMax);
    return demand;
}


// double coinBias[] = { 0, 0.25, 0.5, 0.75, 1.0, -1};
double coinBias[] = { 0, -1};

void performExperiment(WorkData* workData)
{
	unsigned int n, f, ndelta, f_ant, cb, pf;
    // Computes the best number of simulation threads to use:
    // To begin with, it will be equal to the hardware concurrency value given by the C++ runtime. If that is zero (unknown), it will use 1 thread only.
    // Second, it will be not higher than the number of simulations to be run. This is to take into consideration simulations with low repetition counts.
    unsigned int numSimulationThreads = std::min<unsigned int>(std::max<unsigned int>(std::thread::hardware_concurrency(), 1), programArguments.numSimulations);
    double cumulativeCpuTime, estimatedTotalComputeDemand, estimatedConsumedComputeDemand;
    struct rusage ru1, ru2;

    std::cerr << "INFO: using " << numSimulationThreads << " parallel simulation threads.\n\n";

    ndelta = max<unsigned int>(1,(programArguments.nMax - programArguments.nMin) / programArguments.numXPoints);

    estimatedTotalComputeDemand = estimateTotalComputeDemand();
    estimatedConsumedComputeDemand = 0;
    cumulativeCpuTime = 0;
    n = programArguments.nMin;
    getrusage(RUSAGE_SELF, &ru1);
    do {
        cb = 0;
        do {
            pf = programArguments.pfMin;
            f_ant = 0xffffffff;
            do {
                f = (int) ((double) n * pf / 100.0);
                f = std::min<unsigned int>(f, n-(n/2+1));
                if (f != f_ant) {
                    RandomConsensusStatistics consensusStatistics(n, f);
                    ChannelStatistics networkStatistics;
                    thread* simulationThreads[numSimulationThreads];
                    RandomConsensusStatistics* consensusStatisticsAux[numSimulationThreads];
                    ChannelStatistics* networkStatisticsAux[numSimulationThreads];
                    unsigned int numSimulationsPerThread = programArguments.numSimulations / numSimulationThreads;

                    for (unsigned int ti = 0; ti < numSimulationThreads; ti++) {
                        consensusStatisticsAux[ti] = new RandomConsensusStatistics(n, f);
                        networkStatisticsAux[ti] = new ChannelStatistics;
                        simulationThreads[ti] = new thread(simulationThread, n, f, pf, coinBias[cb], numSimulationsPerThread, consensusStatisticsAux[ti], networkStatisticsAux[ti], workData);
                    }
                    for (unsigned int ti = 0; ti < numSimulationThreads; ti++) {
                        simulationThreads[ti]->join();
                        consensusStatistics += *consensusStatisticsAux[ti];
                        networkStatistics += *networkStatisticsAux[ti];
                        delete consensusStatisticsAux[ti];
                        delete networkStatisticsAux[ti];
                        delete simulationThreads[ti];
                    }
                    // TODO: the computation of the average number of messages is not clear: do we divide just by correct executions? Do we substract messages for failed processes in the statistics (would require counting separately per process)? Etc.
                    workData->outputDataFile << n << '\t' << f << '\t' << pf << '\t' << coinBias[cb] << '\t' << networkStatistics.numMessagesSent.average() << '\t' << networkStatistics.numMessagesSent.stdDev() << '\t' << consensusStatistics.numRounds.average() << '\t' << consensusStatistics.numRounds.stdDev() << '\n';
                    workData->outputDataFile << flush;

                    estimatedConsumedComputeDemand += estimateComputeDemand(n,f);
                    getrusage(RUSAGE_SELF, &ru2);
                    cumulativeCpuTime += (ru2.ru_utime.tv_sec + ru2.ru_stime.tv_sec - ru1.ru_utime.tv_sec - ru1.ru_stime.tv_sec + (ru2.ru_utime.tv_usec + ru2.ru_stime.tv_usec - ru1.ru_utime.tv_usec - ru1.ru_stime.tv_usec) / 1e6);
                    ru1 = ru2;
                    if (estimatedConsumedComputeDemand > 0 && programArguments.showProgress && estimatedConsumedComputeDemand / estimatedTotalComputeDemand > 0.01) {
                        double estimatedRemainingCpuTime = (1 - estimatedConsumedComputeDemand / estimatedTotalComputeDemand) / (estimatedConsumedComputeDemand / estimatedTotalComputeDemand) * cumulativeCpuTime;
                        std::cerr << "CPU INFO >>> used: " << cumulativeCpuTime << "s (" << estimatedConsumedComputeDemand / estimatedTotalComputeDemand * 100.0 << "% total estimated), estimated remaining: " << estimatedRemainingCpuTime << "s. Estimated total: " << cumulativeCpuTime + estimatedRemainingCpuTime << "s.\n\n";
                    }

                    f_ant = f;
                }
                pf += programArguments.pfStep;
            } while (pf <= programArguments.pfMax);
            cb++;
        } while (coinBias[cb] >= 0);
        if (n < programArguments.nMax) {
            n += ndelta;
            if (n > programArguments.nMax)
                n = programArguments.nMax;
        }
        else
            n += ndelta;
    } while (n <= programArguments.nMax);
}




int main(int argc, char** argv)
{
    WorkData workData;
//    unsigned int cb;

	readArguments(argc, argv, programArguments);
    workData.outputDataFile.open(programArguments.outputFilename + ".data");
    if (! workData.outputDataFile.is_open()) {
        cerr << "ERROR: cannot create " << programArguments.outputFilename + ".data" << '\n';
        return EX_CANTCREAT;
    }
    workData.outputCommandsFile.open(programArguments.outputFilename + ".r");
    if (! workData.outputCommandsFile.is_open()) {
        cerr << "ERROR: cannot create " << programArguments.outputFilename + ".r" << '\n';
        return EX_CANTCREAT;
    }
    workData.outputExtendedDataFile.open(programArguments.outputFilename + "_extended.data");
    if (! workData.outputExtendedDataFile.is_open()) {
        cerr << "ERROR: cannot create " << programArguments.outputFilename + "_extended.data" << '\n';
        return EX_CANTCREAT;
    }

	srandom(time(NULL) * getpid());

	{
        stringstream s;

        s << "# Experiments for: Random Consensus with Coin\n#\n";
        s << "# n_min = " << programArguments.nMin << '\n';
        s << "# n_max = " << programArguments.nMax << '\n';
        s << "# num_simulations = " << programArguments.numSimulations << '\n';
        s << "# failure_type = " << failureTypeName[programArguments.failureType] << '\n';
        s << "# coin_prob0 = \"" << coinProb0 << "\"\n";
        s << programArguments.comments;
        workData.outputDataFile << s.str();
        workData.outputExtendedDataFile << s.str();
	}

    workData.outputDataFile << "n\tf\tpf\tbias\tnumMessages\tdevMessages\tavgRounds\tdevRounds\n";
    workData.outputExtendedDataFile << "n\tf\tpf\tbias\tnumMessages\tavgRounds\n";

    performExperiment(&workData);

//    cout << "plot(x,z,type=\"o\",xlab=\"Number of processes\",ylab=\"Number of messages\",main=\"Processes - Messages\")\n";
    workData.outputCommandsFile << programArguments.comments;
    workData.outputCommandsFile << "n_min <- " << programArguments.nMin << '\n';
    workData.outputCommandsFile << "n_max <- " << programArguments.nMax << '\n';
    workData.outputCommandsFile << "num_simulations <- " << programArguments.numSimulations << '\n';
    workData.outputCommandsFile << "failure_type <- \"" << failureTypeName[programArguments.failureType] << "\"\n";
    workData.outputCommandsFile << "coin_prob0 <- \"" << coinProb0 << "\"\n";
    workData.outputCommandsFile << "\n\n";
    workData.outputCommandsFile << "df <- read.table(\"" << programArguments.outputFilename + ".data" << "\",header=TRUE)\n";
    workData.outputCommandsFile << "\n\n";
    workData.outputCommandsFile << "summary(df)\n";
    workData.outputCommandsFile << "\n\n";
#ifdef PLOT_MESSAGES
    for (unsigned pf = programArguments.pfMin; pf <= programArguments.pfMax; pf += programArguments.pfStep) {
        workData.outputCommandsFile << "dev.new()\n";
        workData.outputCommandsFile << "plot(numMessages ~ n,data=unique(subset(df,pf==" << pf << " & bias==0)),type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(min(df$numMessages),max(df$numMessages)),main=\"Messages - Processes (f=" << pf << "%)\",xlab=\"# Processes\",ylab=\"# Messages\",col=palette()[0])\n";
        workData.outputCommandsFile << "plotLegend <- \"# Simulations: " << programArguments.numSimulations << ", Failures: " << failureTypeName[programArguments.failureType] << "\";\n";
        workData.outputCommandsFile << "plotLegend <- paste(\"Colors for coin bias:   0=\",palette()[1],sep=\"\")\n";
        cb = 1;
        while (coinBias[cb] >= 0) {
            workData.outputCommandsFile << "plotLegend <- paste(plotLegend,\", " << coinBias[cb] << "=\", palette()[" << cb+1 << "],sep=\"\")\n";
            workData.outputCommandsFile << "par(new=TRUE); par(ann=FALSE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
            workData.outputCommandsFile << "plot(numMessages ~ n,data=unique(subset(df,pf==" << pf << " & bias==" << coinBias[cb] << ")),type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(min(df$numMessages),max(df$numMessages)),col=palette()[" << cb << "])\n";
            cb++;
        }
        workData.outputCommandsFile << "par(new=TRUE); par(ann=TRUE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
        workData.outputCommandsFile << "plot(0,0,type=\"n\",sub=plotLegend,xlab=\"\",ylab=\"\")\n";
    }
#else
#ifdef WITH_COINBIAS
    for (unsigned pf = programArguments.pfMin; pf <= programArguments.pfMax; pf += programArguments.pfStep) {
        workData.outputCommandsFile << "dev.new()\n";
        workData.outputCommandsFile << "plot(avgRounds ~ n,data=unique(subset(df,pf==" << pf << " & bias==0)),type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(min(df$avgRounds),max(df$avgRounds)),main=\"Rounds - Processes (f=" << pf << "%)\",xlab=\"# Processes\",ylab=\"Avg # Rounds\",col=palette()[1])\n";
        workData.outputCommandsFile << "plotLegend <- \"# Simulations: " << programArguments.numSimulations << ", Failures: " << failureTypeName[programArguments.failureType] << "\";\n";
//        workData.outputCommandsFile << "plotLegend <- paste(plotLegend,\"Coin bias:   0=\",palette()[1],sep=\"\")\n";
        cb = 1;
        while (coinBias[cb] >= 0) {
            workData.outputCommandsFile << "plotLegend <- paste(plotLegend,\", " << coinBias[cb] << "=\", palette()[" << cb+1 << "],sep=\"\")\n";
            workData.outputCommandsFile << "par(new=TRUE); par(ann=FALSE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
            workData.outputCommandsFile << "plot(avgRounds ~ n,data=unique(subset(df,pf==" << pf << " & bias==" << coinBias[cb] << ")),type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(min(df$avgRounds),max(df$avgRounds)),col=palette()[" << cb+1 << "])\n";
            cb++;
        }
        workData.outputCommandsFile << "par(new=TRUE); par(ann=TRUE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
        workData.outputCommandsFile << "plot(0,0,type=\"n\",sub=plotLegend,xlab=\"\",ylab=\"\")\n";
    }
#endif

    unsigned int col = 1;
    stringstream dataLegend;
    stringstream plotSubtitle;
    for (unsigned pf = programArguments.pfMin; pf <= programArguments.pfMax; pf += programArguments.pfStep, col++) {
        if (pf == programArguments.pfMin) {
            workData.outputCommandsFile << "dev.new()\n";
            plotSubtitle << "# Simulations: " << programArguments.numSimulations << ", Failures: " << failureTypeName[programArguments.failureType];
            plotSubtitle << ", P(0): " << coinProb0;
            if (programArguments.plotSubtitle.size() > 0)
                plotSubtitle << ", " << programArguments.plotSubtitle;
            workData.outputCommandsFile << "plot(0,type=\"n\",xlim=c(min(df$n),max(df$n)),ylim=c(min(df$avgRounds-df$devRounds/10),max(df$avgRounds+df$devRounds/10)),main=\"Rounds - Processes\",xlab=\"# Processes\",ylab=\"Avg # Rounds\",sub=\"" << plotSubtitle.str() << "\")\n";
        }
        else
            dataLegend << ',';
        workData.outputCommandsFile << "x <- unique(subset(df,pf==" << pf << " & bias==0))$n\n";
        workData.outputCommandsFile << "y <- unique(subset(df,pf==" << pf << " & bias==0))$avgRounds\n";
        workData.outputCommandsFile << "s <- unique(subset(df,pf==" << pf << " & bias==0))$devRounds / 20\n";
        workData.outputCommandsFile << "p <- matrix(c(x,y),ncol=2)\n";
        workData.outputCommandsFile << "sampleWidth <- (max(x) - min(x)) / 100\n";
        workData.outputCommandsFile << "lines(p,col=palette()[" << col << "])\n";
        workData.outputCommandsFile << "points(p,col=palette()[" << col << "])\n";
        dataLegend << "\"pf=" << pf << "\"";
        workData.outputCommandsFile << "segments(x-sampleWidth,y-s,x-sampleWidth,y+s,col=palette()[" << col << "])\n";
        workData.outputCommandsFile << "segments(x+sampleWidth,y-s,x+sampleWidth,y+s,col=palette()[" << col << "])\n";
        workData.outputCommandsFile << "segments(x-sampleWidth,y-s,x+sampleWidth,y-s,col=palette()[" << col << "])\n";
        workData.outputCommandsFile << "segments(x-sampleWidth,y+s,x+sampleWidth,y+s,col=palette()[" << col << "])\n";
    }
    workData.outputCommandsFile << "legend(x=\"topright\",legend=c(" << dataLegend.str() << "),col=palette(),lwd=1)\n";

/*
    unsigned int col = 1;
    for (unsigned pf = programArguments.pfMin; pf <= programArguments.pfMax; pf += programArguments.pfStep, col++) {
        if (pf == programArguments.pfMin) {
            workData.outputCommandsFile << "dev.new()\n";
            workData.outputCommandsFile << "plotLegend <- \"# Simulations: " << programArguments.numSimulations << ", Failures: " << failureTypeName[programArguments.failureType] << "   pf: \";\n";
        }
        else
            workData.outputCommandsFile << "par(new=TRUE); par(ann=FALSE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
        workData.outputCommandsFile << "plotLegend <- paste(plotLegend," << (pf > programArguments.pfMin ? "\", \"," : "" ) << "palette()[" << col << "],\"=" << pf << "\",sep=\"\")\n";
//        workData.outputCommandsFile << "plot(avgRounds ~ n,data=unique(subset(df,pf==" << pf << " & bias==0))," << (pf == programArguments.pfMin ? "main=\"Rounds - Processes\", xlab=\"# Processes\", ylab=\"Avg # Rounds\"," : "" ) << "type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(min(df$avgRounds),max(df$avgRounds)),col=palette()[" << col << "])\n";
        workData.outputCommandsFile << "x <- unique(subset(df,pf==" << pf << " & bias==0))$n\n";
        workData.outputCommandsFile << "y <- unique(subset(df,pf==" << pf << " & bias==0))$avgRounds\n";
        workData.outputCommandsFile << "s <- unique(subset(df,pf==" << pf << " & bias==0))$devRounds / 20\n";
        workData.outputCommandsFile << "sampleWidth <- (max(x) - min(x)) / 100\n";
//        workData.outputCommandsFile << "sampleWidth <- " << std::max<double>((double) (programArguments.nMax - programArguments.nMin) / 100.0, 0.05) << '\n';
        workData.outputCommandsFile << "plot(x,y," << (pf == programArguments.pfMin ? "main=\"Rounds - Processes\", xlab=\"# Processes\", ylab=\"Avg # Rounds\"," : "" ) << "type=\"o\",xlim=c(min(df$n),max(df$n)),ylim=c(min(df$avgRounds-df$devRounds/10),max(df$avgRounds+df$devRounds/10)),col=palette()[" << col << "])\n";
        workData.outputCommandsFile << "segments(x-sampleWidth,y-s,x-sampleWidth,y+s,col=palette()[" << col << "])\n";
        workData.outputCommandsFile << "segments(x+sampleWidth,y-s,x+sampleWidth,y+s,col=palette()[" << col << "])\n";
        workData.outputCommandsFile << "segments(x-sampleWidth,y-s,x+sampleWidth,y-s,col=palette()[" << col << "])\n";
        workData.outputCommandsFile << "segments(x-sampleWidth,y+s,x+sampleWidth,y+s,col=palette()[" << col << "])\n";
    }
    workData.outputCommandsFile << "par(new=TRUE); par(ann=TRUE); par(xaxt=\"n\"); par(yaxt=\"n\");\n";
    workData.outputCommandsFile << "plot(0,0,type=\"n\",sub=plotLegend,xlab=\"\",ylab=\"\")\n";
*/
#endif
    // outputCommandsFile << "persp(n,f,messages,xlab=\"# of processes\",ylab=\"% of faulty processes\",zlab=\"# of messages\",theta=-20,ticktype=\"detailed\",main=\"Processes - Messages\")\n";

    workData.outputCommandsFile << "dev.print(tiff,filename=\"" << programArguments.outputFilename << ".tif\",width=650,compression=\"lzw\",bg=\"white\")\n";
    // convert -background white -alpha remove orig.tif new.tif

    workData.outputDataFile.close();
    workData.outputExtendedDataFile.close();
    workData.outputCommandsFile.close();

    return EX_OK;
}
